# Copyright (c) 2017-2019, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from ...models.neural_network import NeuralNetworkBuilder as _NeuralNetworkBuilder
from ...proto import FeatureTypes_pb2 as _FeatureTypes_pb2
from collections import OrderedDict as _OrderedDict
from ...models import datatypes, _METADATA_VERSION, _METADATA_SOURCE
from ...models import MLModel as _MLModel
from ...models import (
    _MLMODEL_FULL_PRECISION,
    _MLMODEL_HALF_PRECISION,
    _VALID_MLMODEL_PRECISION_TYPES,
)
from ...models.utils import _convert_neural_network_spec_weights_to_fp16

from ..._deps import _HAS_KERAS_TF
from ..._deps import _HAS_KERAS2_TF
from coremltools import __version__ as ct_version

if _HAS_KERAS_TF:
    import keras as _keras
    from . import _layers
    from . import _topology

    _KERAS_LAYER_REGISTRY = {
        _keras.layers.core.Dense: _layers.convert_dense,
        _keras.layers.core.Activation: _layers.convert_activation,
        _keras.layers.advanced_activations.LeakyReLU: _layers.convert_activation,
        _keras.layers.advanced_activations.PReLU: _layers.convert_activation,
        _keras.layers.advanced_activations.ELU: _layers.convert_activation,
        _keras.layers.advanced_activations.ParametricSoftplus: _layers.convert_activation,
        _keras.layers.advanced_activations.ThresholdedReLU: _layers.convert_activation,
        _keras.activations.softmax: _layers.convert_activation,
        _keras.layers.convolutional.Convolution2D: _layers.convert_convolution,
        _keras.layers.convolutional.Deconvolution2D: _layers.convert_convolution,
        _keras.layers.convolutional.AtrousConvolution2D: _layers.convert_convolution,
        _keras.layers.convolutional.AveragePooling2D: _layers.convert_pooling,
        _keras.layers.convolutional.MaxPooling2D: _layers.convert_pooling,
        _keras.layers.pooling.GlobalAveragePooling2D: _layers.convert_pooling,
        _keras.layers.pooling.GlobalMaxPooling2D: _layers.convert_pooling,
        _keras.layers.convolutional.ZeroPadding2D: _layers.convert_padding,
        _keras.layers.convolutional.Cropping2D: _layers.convert_cropping,
        _keras.layers.convolutional.UpSampling2D: _layers.convert_upsample,
        _keras.layers.convolutional.Convolution1D: _layers.convert_convolution1d,
        _keras.layers.convolutional.AtrousConvolution1D: _layers.convert_convolution1d,
        _keras.layers.convolutional.AveragePooling1D: _layers.convert_pooling,
        _keras.layers.convolutional.MaxPooling1D: _layers.convert_pooling,
        _keras.layers.pooling.GlobalAveragePooling1D: _layers.convert_pooling,
        _keras.layers.pooling.GlobalMaxPooling1D: _layers.convert_pooling,
        _keras.layers.convolutional.ZeroPadding1D: _layers.convert_padding,
        _keras.layers.convolutional.Cropping1D: _layers.convert_cropping,
        _keras.layers.convolutional.UpSampling1D: _layers.convert_upsample,
        _keras.layers.recurrent.LSTM: _layers.convert_lstm,
        _keras.layers.recurrent.SimpleRNN: _layers.convert_simple_rnn,
        _keras.layers.recurrent.GRU: _layers.convert_gru,
        _keras.layers.wrappers.Bidirectional: _layers.convert_bidirectional,
        _keras.layers.normalization.BatchNormalization: _layers.convert_batchnorm,
        _keras.engine.topology.Merge: _layers.convert_merge,
        _keras.layers.core.Flatten: _layers.convert_flatten,
        _keras.layers.core.Permute: _layers.convert_permute,
        _keras.layers.core.Reshape: _layers.convert_reshape,
        _keras.layers.embeddings.Embedding: _layers.convert_embedding,
        _keras.layers.core.RepeatVector: _layers.convert_repeat_vector,
        ## All the layers that can be skipped (merged with conv)
        _keras.engine.topology.InputLayer: _layers.default_skip,
        _keras.layers.core.Dropout: _layers.default_skip,
        _keras.layers.wrappers.TimeDistributed: _layers.default_skip,
    }

    _KERAS_SKIP_LAYERS = [
        _keras.layers.core.Dropout,
    ]


def _check_unsupported_layers(model):
    for i, layer in enumerate(model.layers):
        if isinstance(layer, _keras.models.Sequential) or isinstance(
            layer, _keras.models.Model
        ):
            _check_unsupported_layers(layer)
        else:
            if type(layer) not in _KERAS_LAYER_REGISTRY:
                raise ValueError("Keras layer '%s' not supported. " % str(type(layer)))
            if isinstance(layer, _keras.engine.topology.Merge):
                if layer.layers is None:
                    continue
                for merge_layer in layer.layers:
                    if isinstance(merge_layer, _keras.models.Sequential) or isinstance(
                        merge_layer, _keras.models.Model
                    ):
                        _check_unsupported_layers(merge_layer)
            if isinstance(layer, _keras.layers.wrappers.TimeDistributed):
                if type(layer.layer) not in _KERAS_LAYER_REGISTRY:
                    raise ValueError(
                        "Keras layer '%s' not supported. " % str(type(layer.layer))
                    )
            if isinstance(layer, _keras.layers.wrappers.Bidirectional):
                if not isinstance(layer.layer, _keras.layers.recurrent.LSTM):
                    raise ValueError(
                        "Keras bi-directional wrapper conversion supports only "
                        "LSTM layer at this time. "
                    )


def _get_layer_converter_fn(layer):
    """Get the right converter function for Keras
    """
    layer_type = type(layer)
    if layer_type in _KERAS_LAYER_REGISTRY:
        return _KERAS_LAYER_REGISTRY[layer_type]
    else:
        raise TypeError("Keras layer of type %s is not supported." % type(layer))


def _load_keras_model(model_network_path, model_weight_path, custom_objects=None):
    """Load a keras model from disk

    Parameters
    ----------
    model_network_path: str
        Path where the model network path is (json file)

    model_weight_path: str
        Path where the model network weights are (hd5 file)

    custom_objects:
        A dictionary of layers or other custom classes
        or functions used by the model

    Returns
    -------
    model: A keras model
    """
    from keras.models import model_from_json
    import json

    # Load the model network
    json_file = open(model_network_path, "r")
    json_string = json_file.read()
    json_file.close()
    loaded_model_json = json.loads(json_string)

    if not custom_objects:
        custom_objects = {}

    # Load the model weights
    loaded_model = model_from_json(loaded_model_json, custom_objects=custom_objects)
    loaded_model.load_weights(model_weight_path)

    return loaded_model


def _convert(
    model,
    input_names=None,
    output_names=None,
    image_input_names=None,
    is_bgr=False,
    red_bias=0.0,
    green_bias=0.0,
    blue_bias=0.0,
    gray_bias=0.0,
    image_scale=1.0,
    class_labels=None,
    predicted_feature_name=None,
    predicted_probabilities_output="",
    custom_objects=None,
    respect_trainable=False,
):
    if not (_HAS_KERAS_TF):
        raise RuntimeError(
            "keras not found or unsupported version or backend "
            "found. keras conversion API is disabled."
        )
    if isinstance(model, str):
        model = _keras.models.load_model(model, custom_objects=custom_objects)
    elif isinstance(model, tuple):
        model = _load_keras_model(model[0], model[1], custom_objects=custom_objects)

    # Check valid versions
    _check_unsupported_layers(model)

    # Build network graph to represent Keras model
    graph = _topology.NetGraph(model)
    graph.build()
    graph.remove_skip_layers(_KERAS_SKIP_LAYERS)
    graph.insert_1d_permute_layers()
    graph.insert_permute_for_spatial_bn()
    graph.defuse_activation()
    graph.remove_internal_input_layers()
    graph.make_output_layers()

    # The graph should be finalized before executing this
    graph.generate_blob_names()
    graph.add_recurrent_optionals()

    inputs = graph.get_input_layers()
    outputs = graph.get_output_layers()

    # check input / output names validity
    if input_names is not None:
        if isinstance(input_names, str):
            input_names = [input_names]
    else:
        input_names = ["input" + str(i + 1) for i in range(len(inputs))]
    if output_names is not None:
        if isinstance(output_names, str):
            output_names = [output_names]
    else:
        output_names = ["output" + str(i + 1) for i in range(len(outputs))]

    if image_input_names is not None and isinstance(image_input_names, str):
        image_input_names = [image_input_names]

    graph.reset_model_input_names(input_names)
    graph.reset_model_output_names(output_names)

    # Keras -> Core ML input dimension dictionary
    # (None, None) -> [1, 1, 1, 1, 1]
    # (None, D) -> [D] or [D, 1, 1, 1, 1]
    # (None, Seq, D) -> [Seq, 1, D, 1, 1]
    # (None, H, W, C) -> [C, H, W]
    # (D) -> [D]
    # (Seq, D) -> [Seq, 1, 1, D, 1]
    # (Batch, Sequence, D) -> [D]

    # Retrieve input shapes from model
    if type(model.input_shape) is list:
        input_dims = [list(filter(None, x)) for x in model.input_shape]
        unfiltered_shapes = model.input_shape
    else:
        input_dims = [list(filter(None, model.input_shape))]
        unfiltered_shapes = [model.input_shape]

    for idx, dim in enumerate(input_dims):
        unfiltered_shape = unfiltered_shapes[idx]
        if len(dim) == 0:
            # Used to be [None, None] before filtering; indicating unknown
            # sequence length
            input_dims[idx] = tuple([1])
        elif len(dim) == 1:
            s = graph.get_successors(inputs[idx])[0]
            if isinstance(graph.get_keras_layer(s), _keras.layers.embeddings.Embedding):
                # Embedding layer's special input (None, D) where D is actually
                # sequence length
                input_dims[idx] = (1,)
            else:
                input_dims[idx] = dim  # dim is just a number
        elif len(dim) == 2:  # [Seq, D]
            input_dims[idx] = (dim[1],)
        elif len(dim) == 3:  # H,W,C
            if len(unfiltered_shape) > 3:
                # keras uses the reverse notation from us
                input_dims[idx] = (dim[2], dim[0], dim[1])
            else:  # keras provided fixed batch and sequence length, so the input
                # was (batch, sequence, channel)
                input_dims[idx] = (dim[2],)
        else:
            raise ValueError(
                "Input" + input_names[idx] + "has input shape of length" + str(len(dim))
            )

    # Retrieve output shapes from model
    if type(model.output_shape) is list:
        output_dims = [list(filter(None, x)) for x in model.output_shape]
    else:
        output_dims = [list(filter(None, model.output_shape[1:]))]

    for idx, dim in enumerate(output_dims):
        if len(dim) == 1:
            output_dims[idx] = dim
        elif len(dim) == 2:  # [Seq, D]
            output_dims[idx] = (dim[1],)
        elif len(dim) == 3:
            output_dims[idx] = (dim[2], dim[1], dim[0])

    input_types = [datatypes.Array(*dim) for dim in input_dims]
    output_types = [datatypes.Array(*dim) for dim in output_dims]

    # Some of the feature handling is sensitive about string vs. unicode
    input_names = map(str, input_names)
    output_names = map(str, output_names)
    is_classifier = class_labels is not None
    if is_classifier:
        mode = "classifier"
    else:
        mode = None

    # assuming these match
    input_features = list(zip(input_names, input_types))
    output_features = list(zip(output_names, output_types))

    builder = _NeuralNetworkBuilder(input_features, output_features, mode=mode)

    for iter, layer in enumerate(graph.layer_list):
        keras_layer = graph.keras_layer_map[layer]
        print("%d : %s, %s" % (iter, layer, keras_layer))
        if isinstance(keras_layer, _keras.layers.wrappers.TimeDistributed):
            keras_layer = keras_layer.layer
        converter_func = _get_layer_converter_fn(keras_layer)
        input_names, output_names = graph.get_layer_blobs(layer)
        converter_func(builder, layer, input_names, output_names, keras_layer)

    # Set the right inputs and outputs on the model description (interface)
    builder.set_input(input_names, input_dims)
    builder.set_output(output_names, output_dims)

    # Since we aren't mangling anything the user gave us, we only need to update
    # the model interface here
    builder.add_optionals(graph.optional_inputs, graph.optional_outputs)

    # Add classifier classes (if applicable)
    if is_classifier:
        classes_in = class_labels
        if isinstance(classes_in, str):
            import os

            if not os.path.isfile(classes_in):
                raise ValueError(
                    "Path to class labels (%s) does not exist." % classes_in
                )
            with open(classes_in, "r") as f:
                classes = f.read()
            classes = classes.splitlines()
        elif type(classes_in) is list:  # list[int or str]
            classes = classes_in
        else:
            raise ValueError(
                "Class labels must be a list of integers / strings, or a file path"
            )

        if predicted_feature_name is not None:
            builder.set_class_labels(
                classes,
                predicted_feature_name=predicted_feature_name,
                prediction_blob=predicted_probabilities_output,
            )
        else:
            builder.set_class_labels(classes)

    # Set pre-processing paramsters
    builder.set_pre_processing_parameters(
        image_input_names=image_input_names,
        is_bgr=is_bgr,
        red_bias=red_bias,
        green_bias=green_bias,
        blue_bias=blue_bias,
        gray_bias=gray_bias,
        image_scale=image_scale,
    )

    # Return the protobuf spec
    spec = builder.spec
    return spec


def _convert_to_spec(
    model,
    input_names=None,
    output_names=None,
    image_input_names=None,
    input_name_shape_dict={},
    is_bgr=False,
    red_bias=0.0,
    green_bias=0.0,
    blue_bias=0.0,
    gray_bias=0.0,
    image_scale=1.0,
    class_labels=None,
    predicted_feature_name=None,
    model_precision=_MLMODEL_FULL_PRECISION,
    predicted_probabilities_output="",
    add_custom_layers=False,
    custom_conversion_functions=None,
    custom_objects=None,
    input_shapes=None,
    output_shapes=None,
    respect_trainable=False,
    use_float_arraytype=False,
):
    """
    Convert a Keras model to Core ML protobuf specification (.mlmodel).

    Parameters
    ----------
    model: Keras model object | str | (str, str)
        A trained Keras neural network model which can be one of the following:

        - a Keras model object
        - a string with the path to a Keras model file (h5)
        - a tuple of strings, where the first is the path to a Keras model

          architecture (.json file), the second is the path to its weights
          stored in h5 file.

    input_names: [str] | str
        Optional name(s) that can be given to the inputs of the Keras model.
        These names will be used in the interface of the Core ML models to refer
        to the inputs of the Keras model. If not provided, the Keras inputs
        are named to [input1, input2, ..., inputN] in the Core ML model.  When
        multiple inputs are present, the input feature names are in the same
        order as the Keras inputs.

    output_names: [str] | str
        Optional name(s) that can be given to the outputs of the Keras model.
        These names will be used in the interface of the Core ML models to refer
        to the outputs of the Keras model. If not provided, the Keras outputs
        are named to [output1, output2, ..., outputN] in the Core ML model.
        When multiple outputs are present, output feature names are in the same
        order as the Keras inputs.

    image_input_names: [str] | str
        Input names to the Keras model (a subset of the input_names
        parameter) that can be treated as images by Core ML. All other inputs
        are treated as MultiArrays (N-D Arrays).

    input_name_shape_dict: {str: [int]}
        Optional Dictionary of input tensor names and their corresponding shapes expressed
        as a list of ints

    is_bgr: bool | dict()
        Flag indicating the channel order the model internally uses to represent
        color images. Set to True if the internal channel order is BGR,
        otherwise it will be assumed RGB. This flag is applicable only if
        image_input_names is specified. To specify a different value for each
        image input, provide a dictionary with input names as keys.
        Note that this flag is about the models internal channel order.
        An input image can be passed to the model in any color pixel layout
        containing red, green and blue values (e.g. 32BGRA or 32ARGB). This flag
        determines how those pixel values get mapped to the internal multiarray
        representation.

    red_bias: float | dict()
        Bias value to be added to the red channel of the input image.
        Defaults to 0.0
        Applicable only if image_input_names is specified.
        To specify different values for each image input provide a dictionary with input names as keys.

    blue_bias: float | dict()
        Bias value to be added to the blue channel of the input image.
        Defaults to 0.0
        Applicable only if image_input_names is specified.
        To specify different values for each image input provide a dictionary with input names as keys.

    green_bias: float | dict()
        Bias value to be added to the green channel of the input image.
        Defaults to 0.0
        Applicable only if image_input_names is specified.
        To specify different values for each image input provide a dictionary with input names as keys.

    gray_bias: float | dict()
        Bias value to be added to the input image (in grayscale). Defaults
        to 0.0
        Applicable only if image_input_names is specified.
        To specify different values for each image input provide a dictionary with input names as keys.

    image_scale: float | dict()
        Value by which input images will be scaled before bias is added and
        Core ML model makes a prediction. Defaults to 1.0.
        Applicable only if image_input_names is specified.
        To specify different values for each image input provide a dictionary with input names as keys.

    class_labels: list[int or str] | str
        Class labels (applies to classifiers only) that map the index of the
        output of a neural network to labels in a classifier.

        If the provided class_labels is a string, it is assumed to be a
        filepath where classes are parsed as a list of newline separated
        strings.

    predicted_feature_name: str
        Name of the output feature for the class labels exposed in the Core ML
        model (applies to classifiers only). Defaults to 'classLabel'

    model_precision: str
        Precision at which model will be saved. Currently full precision (float) and half precision
        (float16) models are supported. Defaults to '_MLMODEL_FULL_PRECISION' (full precision).

    predicted_probabilities_output: str
        Name of the neural network output to be interpreted as the predicted
        probabilities of the resulting classes. Typically the output of a
        softmax function. Defaults to the first output blob.

    add_custom_layers: bool
        If True, then unknown Keras layer types will be added to the model as
        'custom' layers, which must then be filled in as postprocessing.

    custom_conversion_functions: {'str': (Layer -> CustomLayerParams)}
        A dictionary with keys corresponding to names of custom layers and values
        as functions taking a Keras custom layer and returning a parameter dictionary
        and list of weights.

    custom_objects: {'str': (function)}
        Dictionary that includes a key, value pair of {'<function name>': <function>}
        for custom objects such as custom loss in the Keras model.
        Provide a string of the name of the custom function as a key.
        Provide a function as a value.

    respect_trainable: bool
        If True, then Keras layers that are marked 'trainable' will
        automatically be marked updatable in the Core ML model.

    use_float_arraytype: bool
        If true, the datatype of input/output multiarrays is set to Float32 instead
        of double.

    Returns
    -------
    model: MLModel
        Model in Core ML format.

    Examples
    --------
    .. sourcecode:: python

        # Make a Keras model
        >>> model = Sequential()
        >>> model.add(Dense(num_channels, input_dim = input_dim))

        # Convert it with default input and output names
        >>> import coremltools
        >>> coreml_model = coremltools.converters.keras.convert(model)

        # Saving the Core ML model to a file.
        >>> coreml_model.save('my_model.mlmodel')

    Converting a model with a single image input.

    .. sourcecode:: python

        >>> coreml_model = coremltools.converters.keras.convert(model, input_names =
        ... 'image', image_input_names = 'image')

    Core ML also lets you add class labels to models to expose them as
    classifiers.

    .. sourcecode:: python

        >>> coreml_model = coremltools.converters.keras.convert(model, input_names = 'image',
        ... image_input_names = 'image', class_labels = ['cat', 'dog', 'rat'])

    Class labels for classifiers can also come from a file on disk.

    .. sourcecode:: python

        >>> coreml_model = coremltools.converters.keras.convert(model, input_names =
        ... 'image', image_input_names = 'image', class_labels = 'labels.txt')

    Provide customized input and output names to the Keras inputs and outputs
    while exposing them to Core ML.

    .. sourcecode:: python

        >>> coreml_model = coremltools.converters.keras.convert(model, input_names =
        ...   ['my_input_1', 'my_input_2'], output_names = ['my_output'])

    """
    if model_precision not in _VALID_MLMODEL_PRECISION_TYPES:
        raise RuntimeError("Model precision {} is not valid".format(model_precision))

    if _HAS_KERAS_TF:
        spec = _convert(
            model=model,
            input_names=input_names,
            output_names=output_names,
            image_input_names=image_input_names,
            is_bgr=is_bgr,
            red_bias=red_bias,
            green_bias=green_bias,
            blue_bias=blue_bias,
            gray_bias=gray_bias,
            image_scale=image_scale,
            class_labels=class_labels,
            predicted_feature_name=predicted_feature_name,
            predicted_probabilities_output=predicted_probabilities_output,
            custom_objects=custom_objects,
            respect_trainable=respect_trainable,
        )
    elif _HAS_KERAS2_TF:
        from . import _keras2_converter

        spec = _keras2_converter._convert(
            model=model,
            input_names=input_names,
            output_names=output_names,
            image_input_names=image_input_names,
            input_name_shape_dict=input_name_shape_dict,
            is_bgr=is_bgr,
            red_bias=red_bias,
            green_bias=green_bias,
            blue_bias=blue_bias,
            gray_bias=gray_bias,
            image_scale=image_scale,
            class_labels=class_labels,
            predicted_feature_name=predicted_feature_name,
            predicted_probabilities_output=predicted_probabilities_output,
            add_custom_layers=add_custom_layers,
            custom_conversion_functions=custom_conversion_functions,
            custom_objects=custom_objects,
            input_shapes=input_shapes,
            output_shapes=output_shapes,
            respect_trainable=respect_trainable,
            use_float_arraytype=use_float_arraytype,
        )
    else:
        raise RuntimeError(
            "Keras not found or unsupported version or backend found. keras conversion API is disabled."
        )

    if model_precision == _MLMODEL_HALF_PRECISION and model is not None:
        spec = _convert_neural_network_spec_weights_to_fp16(spec)

    return spec


def convert(
    model,
    input_names=None,
    output_names=None,
    image_input_names=None,
    input_name_shape_dict={},
    is_bgr=False,
    red_bias=0.0,
    green_bias=0.0,
    blue_bias=0.0,
    gray_bias=0.0,
    image_scale=1.0,
    class_labels=None,
    predicted_feature_name=None,
    model_precision=_MLMODEL_FULL_PRECISION,
    predicted_probabilities_output="",
    add_custom_layers=False,
    custom_conversion_functions=None,
    input_shapes=None,
    output_shapes=None,
    respect_trainable=False,
    use_float_arraytype=False,
):
    """
    Convert a Keras model to Core ML protobuf specification (.mlmodel).

    Parameters
    ----------
    model: Keras model object | str | (str, str)

        A trained Keras neural network model which can be one of the following:

        - a Keras model object
        - a string with the path to a Keras model file (h5)
        - a tuple of strings, where the first is the path to a Keras model
    architecture (.json file), the second is the path to its weights stored in h5 file.

    input_names: [str] | str
        Optional name(s) that can be given to the inputs of the Keras model.
        These names will be used in the interface of the Core ML models to refer
        to the inputs of the Keras model. If not provided, the Keras inputs
        are named to [input1, input2, ..., inputN] in the Core ML model.  When
        multiple inputs are present, the input feature names are in the same
        order as the Keras inputs.

    output_names: [str] | str
        Optional name(s) that can be given to the outputs of the Keras model.
        These names will be used in the interface of the Core ML models to refer
        to the outputs of the Keras model. If not provided, the Keras outputs
        are named to [output1, output2, ..., outputN] in the Core ML model.
        When multiple outputs are present, output feature names are in the same
        order as the Keras inputs.

    image_input_names: [str] | str
        Input names to the Keras model (a subset of the input_names
        parameter) that can be treated as images by Core ML. All other inputs
        are treated as MultiArrays (N-D Arrays).

    is_bgr: bool | dict()
        Flag indicating the channel order the model internally uses to represent
        color images. Set to True if the internal channel order is BGR,
        otherwise it will be assumed RGB. This flag is applicable only if
        image_input_names is specified. To specify a different value for each
        image input, provide a dictionary with input names as keys.
        Note that this flag is about the models internal channel order.
        An input image can be passed to the model in any color pixel layout
        containing red, green and blue values (e.g. 32BGRA or 32ARGB). This flag
        determines how those pixel values get mapped to the internal multiarray
        representation.

    red_bias: float | dict()
        Bias value to be added to the red channel of the input image.
        Defaults to 0.0
        Applicable only if image_input_names is specified.
        To specify different values for each image input provide a dictionary with input names as keys.

    blue_bias: float | dict()
        Bias value to be added to the blue channel of the input image.
        Defaults to 0.0
        Applicable only if image_input_names is specified.
        To specify different values for each image input provide a dictionary with input names as keys.

    green_bias: float | dict()
        Bias value to be added to the green channel of the input image.
        Defaults to 0.0
        Applicable only if image_input_names is specified.
        To specify different values for each image input provide a dictionary with input names as keys.

    gray_bias: float | dict()
        Bias value to be added to the input image (in grayscale). Defaults
        to 0.0
        Applicable only if image_input_names is specified.
        To specify different values for each image input provide a dictionary with input names as keys.

    image_scale: float | dict()
        Value by which input images will be scaled before bias is added and
        Core ML model makes a prediction. Defaults to 1.0.
        Applicable only if image_input_names is specified.
        To specify different values for each image input provide a dictionary with input names as keys.

    class_labels: list[int or str] | str
        Class labels (applies to classifiers only) that map the index of the
        output of a neural network to labels in a classifier.

        If the provided class_labels is a string, it is assumed to be a
        filepath where classes are parsed as a list of newline separated
        strings.

    predicted_feature_name: str
        Name of the output feature for the class labels exposed in the Core ML
        model (applies to classifiers only). Defaults to 'classLabel'

    model_precision: str
        Precision at which model will be saved. Currently full precision (float) and half precision
        (float16) models are supported. Defaults to '_MLMODEL_FULL_PRECISION' (full precision).

    predicted_probabilities_output: str
        Name of the neural network output to be interpreted as the predicted
        probabilities of the resulting classes. Typically the output of a
        softmax function. Defaults to the first output blob.

    add_custom_layers: bool
        If yes, then unknown Keras layer types will be added to the model as
        'custom' layers, which must then be filled in as postprocessing.

    custom_conversion_functions: {str:(Layer -> (dict, [weights])) }
        A dictionary with keys corresponding to names of custom layers and values
        as functions taking a Keras custom layer and returning a parameter dictionary
        and list of weights.

    respect_trainable: bool
        If yes, then Keras layers marked 'trainable' will automatically be
        marked updatable in the Core ML model.

    use_float_arraytype: bool
        If true, the datatype of input/output multiarrays is set to Float32 instead
        of double.

    Returns
    -------
    model: MLModel
    Model in Core ML format.

    Examples
    --------
    .. sourcecode:: python

        # Make a Keras model
        >>> model = Sequential()
        >>> model.add(Dense(num_channels, input_dim = input_dim))

        # Convert it with default input and output names
        >>> import coremltools
        >>> coreml_model = coremltools.converters.keras.convert(model)

        # Saving the Core ML model to a file.
        >>> coreml_model.save('my_model.mlmodel')

    Converting a model with a single image input.

    .. sourcecode:: python

        >>> coreml_model = coremltools.converters.keras.convert(model, input_names =
        ... 'image', image_input_names = 'image')

    Core ML also lets you add class labels to models to expose them as
    classifiers.

    .. sourcecode:: python

        >>> coreml_model = coremltools.converters.keras.convert(model, input_names = 'image',
        ... image_input_names = 'image', class_labels = ['cat', 'dog', 'rat'])

    Class labels for classifiers can also come from a file on disk.

    .. sourcecode:: python

        >>> coreml_model = coremltools.converters.keras.convert(model, input_names =
        ... 'image', image_input_names = 'image', class_labels = 'labels.txt')

    Provide customized input and output names to the Keras inputs and outputs
    while exposing them to Core ML.

    .. sourcecode:: python

        >>> coreml_model = coremltools.converters.keras.convert(model, input_names =
        ...   ['my_input_1', 'my_input_2'], output_names = ['my_output'])

    """
    spec = _convert_to_spec(
        model,
        input_names=input_names,
        output_names=output_names,
        image_input_names=image_input_names,
        input_name_shape_dict=input_name_shape_dict,
        is_bgr=is_bgr,
        red_bias=red_bias,
        green_bias=green_bias,
        blue_bias=blue_bias,
        gray_bias=gray_bias,
        image_scale=image_scale,
        class_labels=class_labels,
        predicted_feature_name=predicted_feature_name,
        model_precision=model_precision,
        predicted_probabilities_output=predicted_probabilities_output,
        add_custom_layers=add_custom_layers,
        custom_conversion_functions=custom_conversion_functions,
        input_shapes=input_shapes,
        output_shapes=output_shapes,
        respect_trainable=respect_trainable,
        use_float_arraytype=use_float_arraytype,
    )

    model = _MLModel(spec)

    from keras import __version__ as keras_version

    model.user_defined_metadata[_METADATA_VERSION] = ct_version
    model.user_defined_metadata[_METADATA_SOURCE] = "keras=={0}".format(keras_version)

    return model



# -*- coding: utf-8 -*-

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from __future__ import print_function as _
from __future__ import division as _
from __future__ import absolute_import as _
from coremltools.converters.mil.mil import types
from .tfssa import ParsedNode


class ParsedTFNode(ParsedNode):
    """
    A parsed TensorFlow Node.

    name: The name of the node (str)
    op: The operation represented by the node (str)
    datatype: The type of the node. (type)
    value: The value of the node if available
    inputs: The list of nodes which are inputs to this node (list[str])
    control_inputs: The list of nodes which have to be executed before this node (list[str])
    attr: The attributes of the node
    outputs: The list of nodes which consume the result of this node (list[str])
    control_outputs: The list of nodes which have to be executed after this node (list[str])
    """

    def __init__(self, tfnode=None):
        super(ParsedTFNode, self).__init__()
        self.original_node = tfnode

        if tfnode is not None:
            from .parse import parse_attr

            self.name = tfnode.name
            if tfnode.op == "PlaceholderWithDefault":
                self.op = "Placeholder"
            else:
                self.op = tfnode.op
            self.inputs = [x for x in tfnode.input if not x.startswith("^")]
            self.control_inputs = [x[1:] for x in tfnode.input if x.startswith("^")]
            self.attr = {k: parse_attr(v) for k, v in tfnode.attr.items()}

    def parse_from_attr(self):
        if "value" in self.attr:
            self.datatype = self.attr["value"].__class__
        elif "_output_shapes" in self.attr:
            output_shapes = self.attr["_output_shapes"]
            if output_shapes[0] is not None and len(output_shapes[0]) > 0:
                if "dtype" in self.attr:
                    rettype = types.tensor(self.attr["dtype"], tuple(output_shapes[0]))
                elif "T" in self.attr:
                    rettype = types.tensor(self.attr["T"], tuple(output_shapes[0]))
                elif "Tparams" in self.attr:
                    rettype = types.tensor(
                        self.attr["Tparams"], tuple(output_shapes[0])
                    )
                else:
                    raise NotImplementedError(
                        "Op-(%s) %s not implemented\nWith attribute:"
                        + str(self.attr) % (self.op, self.name)
                    )
                self.datatype = rettype
            elif "dtype" in self.attr:
                self.datatype = self.attr["dtype"]
        elif "shape" in self.attr:
            shape = self.attr["shape"]
            assert "dtype" in self.attr
            if len(shape) == 0:
                self.datatype = self.attr["dtype"]
            else:
                self.datatype = types.tensor(self.attr["dtype"], shape)
        elif "dtype" in self.attr:
            self.datatype = self.attr["dtype"]

    def _copy_impl(self, dest):
        dest = super(ParsedTFNode, self)._copy_impl(dest)
        dest.original_node = self.original_node
        return dest

    def __copy__(self):
        return self._copy_impl(ParsedTFNode())



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from coremltools.converters.mil import testing_reqs
from coremltools.converters.mil.testing_reqs import *

from .testing_utils import UNK_SYM, run_compare_builder

backends = testing_reqs.backends


class TestSliceByIndex:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        x_val = np.array(list(range(24))).reshape((2, 3, 4)).astype(np.float32)
        begin_val = np.array([1, 1, 1], dtype=np.int32)
        end_val = np.array([2, 3, 3], dtype=np.int32)
        input_placeholders = {
            "x": mb.placeholder(shape=x_val.shape),
            "begin": mb.placeholder(shape=begin_val.shape, dtype=types.int32),
            "end": mb.placeholder(shape=end_val.shape, dtype=types.int32),
        }
        input_values = {"x": x_val, "begin": begin_val, "end": end_val}

        def build(x, begin, end):
            begin_c = mb.const(val=begin_val)
            end_c = mb.const(val=end_val)
            return [
                mb.slice_by_index(x=x, begin=begin, end=end),
                mb.slice_by_index(x=x, begin=begin_c, end=end_c)
            ]

        expected_output_types = [(UNK_SYM, UNK_SYM, UNK_SYM, types.fp32)] * 2
        expected_outputs = [np.array([[[17, 18], [21, 22]]], dtype=np.float32)] * 2
        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @ssa_fn
    def test_builder_eval(self):
        x_val = np.array(list(range(24))).reshape((2, 3, 4))
        v = [
            mb.slice_by_index(x=x_val, begin=[1, 1, 1], end=[2, 2, 2]),
            mb.slice_by_index(
                x=x_val, begin=[1, 1, 1], end=[2, 3, 4], stride=[1, 1, 2]
            ),
            mb.slice_by_index(x=x_val, begin=[-1, -3, -3], end=[-1, -1, -1]),
            mb.slice_by_index(x=x_val, begin=[0, 0, -3], end=[-1, -2, -2]),
            mb.slice_by_index(
                x=x_val, begin=[-1, -1, -1], end=[0, 1, -3], stride=[-2, -1, -3]
            ),
            mb.slice_by_index(
                x=x_val,
                begin=[1, 1, 1],
                end=[2, 3, 4],
                stride=[1, 1, 2],
                begin_mask=[True, False, True],
            ),
            mb.slice_by_index(
                x=x_val,
                begin=[1, 1, 1],
                end=[2, 3, 4],
                stride=[1, 1, 2],
                begin_mask=[True, False, True],
                end_mask=[True, True, False],
            ),
            mb.slice_by_index(
                x=x_val,
                begin=[1, 1, 1],
                end=[2, 3, 4],
                stride=[1, 1, 2],
                begin_mask=[False, False, True],
                end_mask=[True, False, False],
                squeeze_mask=[False, True, False],
            ),
            mb.slice_by_index(
                x=x_val,
                begin=[0, 0, 0],
                end=[0, 0, 0],
                stride=[1, 1, 1],
                begin_mask=[True, True, True],
                end_mask=[True, True, True],
            ),
            mb.slice_by_index(
                x=x_val,
                begin=[1, 1, 1],
                end=[2, 2, 0],
                stride=[1, 1, 1],
                squeeze_mask=[False, False, True],
            ),
            mb.slice_by_index(
                x=x_val,
                begin=[1, 0, 0],
                end=[2, 0, 0],
                stride=[1, 1, 1],
                begin_mask=[False, True, True],
                end_mask=[False, True, True],
            ),
            mb.slice_by_index(
                x=x_val,
                begin=[0, 0, 0],
                end=[0, 0, 0],
                stride=[1, 1, 1],
                begin_mask=[True, True, True],
                end_mask=[True, True, True],
            ),
            mb.slice_by_index(
                x=x_val,
                begin=[1, 0, 1],
                end=[2, 0, 2],
                stride=[1, 1, 1],
                begin_mask=[False, True, False],
                end_mask=[False, True, False],
            ),
            mb.slice_by_index(
                x=x_val,
                begin=[0, 0, 1],
                end=[0, 0, 0],
                stride=[1, 1, 1],
                begin_mask=[True, True, False],
                end_mask=[True, True, False],
                squeeze_mask=[False, False, True],
            ),
            mb.slice_by_index(
                x=x_val,
                begin=[0, 0, 0],
                end=[0, 0, 0],
                stride=[1, 1, 1],
                begin_mask=[False, False, True],
                end_mask=[False, False, True],
                squeeze_mask=[True, True, False],
            ),
            mb.slice_by_index(
                x=x_val,
                begin=[1, 0, 0],
                end=[2, 0, 0],
                stride=[1, 1, 1],
                begin_mask=[False, True, True],
                end_mask=[False, True, True],
            ),
            mb.slice_by_index(
                x=x_val,
                begin=[1, 1, 0],
                end=[2, 2, 0],
                stride=[1, 1, 1],
                begin_mask=[False, False, True],
                end_mask=[False, False, True],
            ),
            mb.slice_by_index(
                x=x_val,
                begin=[1, 0, 0],
                end=[0, 0, 0],
                stride=[1, 1, 1],
                begin_mask=[False, True, True],
                end_mask=[False, True, True],
                squeeze_mask=[True, False, False],
            ),
            mb.slice_by_index(
                x=x_val,
                begin=[0, 0, 0],
                end=[0, 0, 0],
                begin_mask=[True, True, True],
                end_mask=[True, True, True],
            ),
            mb.slice_by_index(
                x=x_val,
                begin=[0, 0, 0],
                end=[0, 0, 0],
                stride=[1, 1, -1],
                begin_mask=[True, True, True],
                end_mask=[True, True, True],
            ),
        ]
        ans = [
            x_val[1:2, 1:2, 1:2],
            x_val[1:2, 1:3, 1:4:2],
            x_val[-3:-1, -3:-1, -3:-1],
            x_val[0:-1, 0:-2, -3:-2],
            x_val[-1:0:-2, -1:1:-1, -1:-3:-3],
            x_val[:2, 1:3, :4:2],
            x_val[:, 1:, :4:2],
            x_val[1::1, 1, :3:2],
            x_val[:, :, :],
            x_val[1:2, 1:2, 1],
            x_val[1:2, ...],
            x_val[...],
            x_val[1:2, ..., 1:2],
            x_val[..., 1],
            x_val[0, 0, :],
            x_val[1:2],
            x_val[1:2, 1:2],
            x_val[1],
            x_val[:],
            x_val[..., ::-1],
        ]
        for idx in range(len(v)):
            assert is_close(ans[idx], v[idx].val)



# -*- coding: utf-8 -*-
from __future__ import print_function as _
from __future__ import division as _
from __future__ import absolute_import as _

from coremltools.converters.mil.mil.passes.pass_registry import register_pass
from coremltools.converters.mil.mil import Builder as mb
import numpy as np


def _remove_elementwise_binary(op, block, x, y):
    # We remove the ops that has op.x == x or op.y == y
    if x is not None and op.x.val is not None and np.all(op.x.val == x):
        input_var = op.y
        input_op = input_var.op
    elif y is not None and op.y.val is not None and np.all(op.y.val == y):
        input_var = op.x
        input_op = input_var.op
    else:
        return False

    input_shape = input_var.sym_type
    output_shape = op.outputs[0].sym_type

    # We might be using elementwise as broadcasting
    if input_shape != output_shape:
        return False

    op.enclosing_block.replace_uses_of_var_after_op(
        anchor_op=input_op, old_var=op.outputs[0], new_var=input_var
    )
    block.remove_ops([op])

    return True


def remove_elementwise(op, block):

    if op.op_type in {"add"}:
        return _remove_elementwise_binary(op, block, 0, 0)
    elif op.op_type in {"mul"}:
        return _remove_elementwise_binary(op, block, 1, 1)
    elif op.op_type in {"floor_div", "pow", "real_div"}:
        return _remove_elementwise_binary(op, block, None, 1)
    elif op.op_type in {"sub"}:
        return _remove_elementwise_binary(op, block, None, 0)
    else:
        return False


def remove_same_shape(op, block):
    input_shape = op.x.sym_type
    output_shape = op.outputs[0].sym_type

    if input_shape != output_shape:
        return False

    input_var = op.x
    input_op = input_var.op

    op.enclosing_block.replace_uses_of_var_after_op(
        anchor_op=input_op, old_var=op.outputs[0], new_var=input_var
    )

    # Remove all the ops at once
    block.remove_ops([op])
    return True


def remove_linear(op, block):
    if op.alpha.val != 1 or op.beta.val != 0:
        return False

    input_var = op.x
    input_op = input_var.op

    op.enclosing_block.replace_uses_of_var_after_op(
        anchor_op=input_op, old_var=op.outputs[0], new_var=input_var
    )

    # Remove all the ops at once
    block.remove_ops([op])
    return True


_SUPPORTED_OPS = {
    "add",
    "mul",
    "floor_div",
    "pow",
    "real_div",
    "sub",
    "reshape",
    "split",
    "slice_by_index",
    "slice_by_size",
    "pad",
    "tile",
    "upsample_nearest_neighbor",
    "upsample_bilinear",
    "resize_bilinear",
    "crop",
    "linear_activation"
}
op_to_removal_fn = {
    "add": remove_elementwise,
    "mul": remove_elementwise,
    "floor_div": remove_elementwise,
    "pow": remove_elementwise,
    "real_div": remove_elementwise,
    "sub": remove_elementwise,
    "reshape": remove_same_shape,
    "split": remove_same_shape,
    "slice_by_index": remove_same_shape,
    "slice_by_size": remove_same_shape,
    "pad": remove_same_shape,
    "tile": remove_same_shape,
    "upsample_nearest_neighbor": remove_same_shape,
    "upsample_bilinear": remove_same_shape,
    "resize_bilinear": remove_same_shape,
    "crop": remove_same_shape,
    "linear_activation": remove_linear,
}


def match_pattern(op):
    # abort if op output is a block output
    if op.outputs[0] in op.enclosing_block.outputs:
        return None

    if op.op_type in _SUPPORTED_OPS:

        if len(op.outputs) != 1:
            return None
        return op_to_removal_fn[op.op_type]

    return None


def noop_elimination_block(block):
    for op in list(block.operations):
        for b in op.blocks:
            block_changed = True
            while block_changed:
                block_changed = noop_elimination_block(b)
        if len(op.blocks) > 0:
            continue

        remove_fn = match_pattern(op)
        if remove_fn is not None:
            with block:
                status = remove_fn(op, block)
            # has to break as the downstream iterator is affected.
            if status:
                return status
    return False


@register_pass(namespace="common")
def noop_elimination(prog):
    """
    We remove ops that has no effect.

    Given:
        %1 (1, 96, 128, 64, fp32) = ...
        %2 (1, 96, 128, 64, fp32) = reshape(%1)
        ...
        %3 (1, 96, 128, 64, fp32) = add(%2, constant)
        ...

    Result:
        %1 (1, 96, 128, 64, fp32) = ...
        %3 (1, 96, 128, 64, fp32) = add(%1, constant)
        ...

    """
    for f_name, f in prog.functions.items():
        block_changed = True
        while block_changed:
            block_changed = noop_elimination_block(f)



# -*- coding: utf-8 -*-

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from __future__ import print_function as _
from __future__ import division as _
from __future__ import absolute_import as _

from coremltools.converters.mil.mil.passes.pass_registry import register_pass
from coremltools.converters.mil.mil import Builder as mb
import numpy as np


def match_pattern(op):
    if op.outputs[0] in op.enclosing_block.outputs:
        return None

    if op.op_type == "mul":
        # find add
        child_ops = op.outputs[0].child_ops
        if len(child_ops) == 1:
            add_op_candidate = list(child_ops)[0]
            if add_op_candidate.op_type == "add":
                return add_op_candidate
    return None


def _find_const_input_val(op):
    if op.x.val is not None:
        return op.x.val
    if op.y.val is not None:
        return op.y.val
    return None


def _check_shape(arr):
    """
    return True if shape is of form
    (1,C,1,1) or (C,1,1)
    """
    rank = len(arr.shape)
    if not (rank == 3 or rank == 4):
        return False
    C = arr.shape[-3]
    if not (arr.shape == (1, C, 1, 1) or arr.shape == (C, 1, 1)):
        return False
    return True


def try_to_transform(mul_op, add_op, block):
    non_const_input_mul = mul_op.x if mul_op.x.val is None else mul_op.y
    if non_const_input_mul.rank != 4:
        return False

    gamma = _find_const_input_val(mul_op)
    beta = _find_const_input_val(add_op)
    if gamma is None or beta is None:
        return False

    if not (isinstance(gamma, np.ndarray) and isinstance(beta, np.ndarray)):
        return False

    # check that gamma and beta have shape (1,C,1,1) or (C,1,1)
    # that is they are doing vector addition on the axis=-3, which is what the
    # batchnorm layer does (batchnorm layer only works on rank 4 input tensors)
    if not (_check_shape(gamma) and _check_shape(beta)):
        return False

    C = gamma.shape[-3]
    if C == 1:
        return False

    out_name = add_op.outputs[0].name
    x = mb.batch_norm(
        x=non_const_input_mul,
        mean=np.zeros((C,), np.float32),
        variance=np.ones((C,), np.float32),
        gamma=np.squeeze(gamma),
        beta=np.squeeze(beta),
        name=out_name,
        before_op=mul_op,
    )

    add_op.enclosing_block.replace_uses_of_var_after_op(
        anchor_op=add_op, old_var=add_op.outputs[0], new_var=x
    )
    # Remove all the ops at once
    block.remove_ops([mul_op, add_op])
    return True


def fuse_elementwise_to_batchnorm_block(block):
    fusion_status = False
    for op in list(block.operations):
        for b in op.blocks:
            block_changed = True
            while block_changed:
                block_changed = fuse_elementwise_to_batchnorm_block(b)
        if len(op.blocks) > 0:
            # This op can't be mul
            continue

        add_op = match_pattern(op)
        if add_op is not None:
            with block:
                fusion_status = try_to_transform(op, add_op, block)
            # has to break as the downstream iterator is affected.
            if fusion_status:
                return fusion_status
    return fusion_status


@register_pass(namespace="common")
def fuse_elementwise_to_batchnorm(prog):
    """
    Fold mul + add into a batch norm,
    if the const feeding into the mul/add is of shape (1,C,1,1) or (C,1,1)
    and input to mul is of rank 4.

    Given:
             [Const]   [Const]
                |         |
                V         V
    [...] --> [Mul] --> [Add] --> [...]

    That is,

        %2 = op1(%1)
        %3 = mul(%2, constant)
        %4 = add(%3, constant)
        %5 = op2(%4)
        ...

    Result:

    [...] --> [BatchNorm] --> [...]

    That is,
        %2 = op1(%1)
        %4 = batchnorm(%2)
        %5 = op2(%4)
        ...

    """
    for f_name, f in prog.functions.items():
        block_changed = True
        while block_changed:
            block_changed = fuse_elementwise_to_batchnorm_block(f)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: FeatureVectorizer.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='FeatureVectorizer.proto',
  package='CoreML.Specification',
  syntax='proto3',
  serialized_pb=_b('\n\x17\x46\x65\x61tureVectorizer.proto\x12\x14\x43oreML.Specification\"\x98\x01\n\x11\x46\x65\x61tureVectorizer\x12\x46\n\tinputList\x18\x01 \x03(\x0b\x32\x33.CoreML.Specification.FeatureVectorizer.InputColumn\x1a;\n\x0bInputColumn\x12\x13\n\x0binputColumn\x18\x01 \x01(\t\x12\x17\n\x0finputDimensions\x18\x02 \x01(\x04\x42\x02H\x03\x62\x06proto3')
)




_FEATUREVECTORIZER_INPUTCOLUMN = _descriptor.Descriptor(
  name='InputColumn',
  full_name='CoreML.Specification.FeatureVectorizer.InputColumn',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='inputColumn', full_name='CoreML.Specification.FeatureVectorizer.InputColumn.inputColumn', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='inputDimensions', full_name='CoreML.Specification.FeatureVectorizer.InputColumn.inputDimensions', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=143,
  serialized_end=202,
)

_FEATUREVECTORIZER = _descriptor.Descriptor(
  name='FeatureVectorizer',
  full_name='CoreML.Specification.FeatureVectorizer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='inputList', full_name='CoreML.Specification.FeatureVectorizer.inputList', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_FEATUREVECTORIZER_INPUTCOLUMN, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=50,
  serialized_end=202,
)

_FEATUREVECTORIZER_INPUTCOLUMN.containing_type = _FEATUREVECTORIZER
_FEATUREVECTORIZER.fields_by_name['inputList'].message_type = _FEATUREVECTORIZER_INPUTCOLUMN
DESCRIPTOR.message_types_by_name['FeatureVectorizer'] = _FEATUREVECTORIZER
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

FeatureVectorizer = _reflection.GeneratedProtocolMessageType('FeatureVectorizer', (_message.Message,), dict(

  InputColumn = _reflection.GeneratedProtocolMessageType('InputColumn', (_message.Message,), dict(
    DESCRIPTOR = _FEATUREVECTORIZER_INPUTCOLUMN,
    __module__ = 'FeatureVectorizer_pb2'
    # @@protoc_insertion_point(class_scope:CoreML.Specification.FeatureVectorizer.InputColumn)
    ))
  ,
  DESCRIPTOR = _FEATUREVECTORIZER,
  __module__ = 'FeatureVectorizer_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.FeatureVectorizer)
  ))
_sym_db.RegisterMessage(FeatureVectorizer)
_sym_db.RegisterMessage(FeatureVectorizer.InputColumn)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('H\003'))
# @@protoc_insertion_point(module_scope)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: FeatureTypes.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='FeatureTypes.proto',
  package='CoreML.Specification',
  syntax='proto3',
  serialized_pb=_b('\n\x12\x46\x65\x61tureTypes.proto\x12\x14\x43oreML.Specification\"\x12\n\x10Int64FeatureType\"\x13\n\x11\x44oubleFeatureType\"\x13\n\x11StringFeatureType\"3\n\tSizeRange\x12\x12\n\nlowerBound\x18\x01 \x01(\x04\x12\x12\n\nupperBound\x18\x02 \x01(\x03\"\xfe\x04\n\x10ImageFeatureType\x12\r\n\x05width\x18\x01 \x01(\x03\x12\x0e\n\x06height\x18\x02 \x01(\x03\x12V\n\x0f\x65numeratedSizes\x18\x15 \x01(\x0b\x32;.CoreML.Specification.ImageFeatureType.EnumeratedImageSizesH\x00\x12O\n\x0eimageSizeRange\x18\x1f \x01(\x0b\x32\x35.CoreML.Specification.ImageFeatureType.ImageSizeRangeH\x00\x12\x45\n\ncolorSpace\x18\x03 \x01(\x0e\x32\x31.CoreML.Specification.ImageFeatureType.ColorSpace\x1a*\n\tImageSize\x12\r\n\x05width\x18\x01 \x01(\x04\x12\x0e\n\x06height\x18\x02 \x01(\x04\x1aW\n\x14\x45numeratedImageSizes\x12?\n\x05sizes\x18\x01 \x03(\x0b\x32\x30.CoreML.Specification.ImageFeatureType.ImageSize\x1a{\n\x0eImageSizeRange\x12\x33\n\nwidthRange\x18\x01 \x01(\x0b\x32\x1f.CoreML.Specification.SizeRange\x12\x34\n\x0bheightRange\x18\x02 \x01(\x0b\x32\x1f.CoreML.Specification.SizeRange\"F\n\nColorSpace\x12\x17\n\x13INVALID_COLOR_SPACE\x10\x00\x12\r\n\tGRAYSCALE\x10\n\x12\x07\n\x03RGB\x10\x14\x12\x07\n\x03\x42GR\x10\x1e\x42\x11\n\x0fSizeFlexibility\"\x8e\x05\n\x10\x41rrayFeatureType\x12\r\n\x05shape\x18\x01 \x03(\x03\x12\x46\n\x08\x64\x61taType\x18\x02 \x01(\x0e\x32\x34.CoreML.Specification.ArrayFeatureType.ArrayDataType\x12S\n\x10\x65numeratedShapes\x18\x15 \x01(\x0b\x32\x37.CoreML.Specification.ArrayFeatureType.EnumeratedShapesH\x00\x12G\n\nshapeRange\x18\x1f \x01(\x0b\x32\x31.CoreML.Specification.ArrayFeatureType.ShapeRangeH\x00\x12\x19\n\x0fintDefaultValue\x18) \x01(\x05H\x01\x12\x1b\n\x11\x66loatDefaultValue\x18\x33 \x01(\x02H\x01\x12\x1c\n\x12\x64oubleDefaultValue\x18= \x01(\x01H\x01\x1a\x16\n\x05Shape\x12\r\n\x05shape\x18\x01 \x03(\x03\x1aP\n\x10\x45numeratedShapes\x12<\n\x06shapes\x18\x01 \x03(\x0b\x32,.CoreML.Specification.ArrayFeatureType.Shape\x1a\x41\n\nShapeRange\x12\x33\n\nsizeRanges\x18\x01 \x03(\x0b\x32\x1f.CoreML.Specification.SizeRange\"V\n\rArrayDataType\x12\x1b\n\x17INVALID_ARRAY_DATA_TYPE\x10\x00\x12\r\n\x07\x46LOAT32\x10\xa0\x80\x04\x12\x0c\n\x06\x44OUBLE\x10\xc0\x80\x04\x12\x0b\n\x05INT32\x10\xa0\x80\x08\x42\x12\n\x10ShapeFlexibilityB\x16\n\x14\x64\x65\x66\x61ultOptionalValue\"\xa4\x01\n\x15\x44ictionaryFeatureType\x12>\n\x0cint64KeyType\x18\x01 \x01(\x0b\x32&.CoreML.Specification.Int64FeatureTypeH\x00\x12@\n\rstringKeyType\x18\x02 \x01(\x0b\x32\'.CoreML.Specification.StringFeatureTypeH\x00\x42\t\n\x07KeyType\"\xcd\x01\n\x13SequenceFeatureType\x12;\n\tint64Type\x18\x01 \x01(\x0b\x32&.CoreML.Specification.Int64FeatureTypeH\x00\x12=\n\nstringType\x18\x03 \x01(\x0b\x32\'.CoreML.Specification.StringFeatureTypeH\x00\x12\x32\n\tsizeRange\x18\x65 \x01(\x0b\x32\x1f.CoreML.Specification.SizeRangeB\x06\n\x04Type\"\xee\x03\n\x0b\x46\x65\x61tureType\x12;\n\tint64Type\x18\x01 \x01(\x0b\x32&.CoreML.Specification.Int64FeatureTypeH\x00\x12=\n\ndoubleType\x18\x02 \x01(\x0b\x32\'.CoreML.Specification.DoubleFeatureTypeH\x00\x12=\n\nstringType\x18\x03 \x01(\x0b\x32\'.CoreML.Specification.StringFeatureTypeH\x00\x12;\n\timageType\x18\x04 \x01(\x0b\x32&.CoreML.Specification.ImageFeatureTypeH\x00\x12@\n\x0emultiArrayType\x18\x05 \x01(\x0b\x32&.CoreML.Specification.ArrayFeatureTypeH\x00\x12\x45\n\x0e\x64ictionaryType\x18\x06 \x01(\x0b\x32+.CoreML.Specification.DictionaryFeatureTypeH\x00\x12\x41\n\x0csequenceType\x18\x07 \x01(\x0b\x32).CoreML.Specification.SequenceFeatureTypeH\x00\x12\x13\n\nisOptional\x18\xe8\x07 \x01(\x08\x42\x06\n\x04TypeB\x02H\x03\x62\x06proto3')
)



_IMAGEFEATURETYPE_COLORSPACE = _descriptor.EnumDescriptor(
  name='ColorSpace',
  full_name='CoreML.Specification.ImageFeatureType.ColorSpace',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='INVALID_COLOR_SPACE', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GRAYSCALE', index=1, number=10,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RGB', index=2, number=20,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BGR', index=3, number=30,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=709,
  serialized_end=779,
)
_sym_db.RegisterEnumDescriptor(_IMAGEFEATURETYPE_COLORSPACE)

_ARRAYFEATURETYPE_ARRAYDATATYPE = _descriptor.EnumDescriptor(
  name='ArrayDataType',
  full_name='CoreML.Specification.ArrayFeatureType.ArrayDataType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='INVALID_ARRAY_DATA_TYPE', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FLOAT32', index=1, number=65568,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DOUBLE', index=2, number=65600,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INT32', index=3, number=131104,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=1325,
  serialized_end=1411,
)
_sym_db.RegisterEnumDescriptor(_ARRAYFEATURETYPE_ARRAYDATATYPE)


_INT64FEATURETYPE = _descriptor.Descriptor(
  name='Int64FeatureType',
  full_name='CoreML.Specification.Int64FeatureType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=44,
  serialized_end=62,
)


_DOUBLEFEATURETYPE = _descriptor.Descriptor(
  name='DoubleFeatureType',
  full_name='CoreML.Specification.DoubleFeatureType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=64,
  serialized_end=83,
)


_STRINGFEATURETYPE = _descriptor.Descriptor(
  name='StringFeatureType',
  full_name='CoreML.Specification.StringFeatureType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=85,
  serialized_end=104,
)


_SIZERANGE = _descriptor.Descriptor(
  name='SizeRange',
  full_name='CoreML.Specification.SizeRange',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='lowerBound', full_name='CoreML.Specification.SizeRange.lowerBound', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='upperBound', full_name='CoreML.Specification.SizeRange.upperBound', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=106,
  serialized_end=157,
)


_IMAGEFEATURETYPE_IMAGESIZE = _descriptor.Descriptor(
  name='ImageSize',
  full_name='CoreML.Specification.ImageFeatureType.ImageSize',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='width', full_name='CoreML.Specification.ImageFeatureType.ImageSize.width', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='height', full_name='CoreML.Specification.ImageFeatureType.ImageSize.height', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=451,
  serialized_end=493,
)

_IMAGEFEATURETYPE_ENUMERATEDIMAGESIZES = _descriptor.Descriptor(
  name='EnumeratedImageSizes',
  full_name='CoreML.Specification.ImageFeatureType.EnumeratedImageSizes',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='sizes', full_name='CoreML.Specification.ImageFeatureType.EnumeratedImageSizes.sizes', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=495,
  serialized_end=582,
)

_IMAGEFEATURETYPE_IMAGESIZERANGE = _descriptor.Descriptor(
  name='ImageSizeRange',
  full_name='CoreML.Specification.ImageFeatureType.ImageSizeRange',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='widthRange', full_name='CoreML.Specification.ImageFeatureType.ImageSizeRange.widthRange', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='heightRange', full_name='CoreML.Specification.ImageFeatureType.ImageSizeRange.heightRange', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=584,
  serialized_end=707,
)

_IMAGEFEATURETYPE = _descriptor.Descriptor(
  name='ImageFeatureType',
  full_name='CoreML.Specification.ImageFeatureType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='width', full_name='CoreML.Specification.ImageFeatureType.width', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='height', full_name='CoreML.Specification.ImageFeatureType.height', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='enumeratedSizes', full_name='CoreML.Specification.ImageFeatureType.enumeratedSizes', index=2,
      number=21, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='imageSizeRange', full_name='CoreML.Specification.ImageFeatureType.imageSizeRange', index=3,
      number=31, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='colorSpace', full_name='CoreML.Specification.ImageFeatureType.colorSpace', index=4,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_IMAGEFEATURETYPE_IMAGESIZE, _IMAGEFEATURETYPE_ENUMERATEDIMAGESIZES, _IMAGEFEATURETYPE_IMAGESIZERANGE, ],
  enum_types=[
    _IMAGEFEATURETYPE_COLORSPACE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='SizeFlexibility', full_name='CoreML.Specification.ImageFeatureType.SizeFlexibility',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=160,
  serialized_end=798,
)


_ARRAYFEATURETYPE_SHAPE = _descriptor.Descriptor(
  name='Shape',
  full_name='CoreML.Specification.ArrayFeatureType.Shape',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='shape', full_name='CoreML.Specification.ArrayFeatureType.Shape.shape', index=0,
      number=1, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1152,
  serialized_end=1174,
)

_ARRAYFEATURETYPE_ENUMERATEDSHAPES = _descriptor.Descriptor(
  name='EnumeratedShapes',
  full_name='CoreML.Specification.ArrayFeatureType.EnumeratedShapes',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='shapes', full_name='CoreML.Specification.ArrayFeatureType.EnumeratedShapes.shapes', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1176,
  serialized_end=1256,
)

_ARRAYFEATURETYPE_SHAPERANGE = _descriptor.Descriptor(
  name='ShapeRange',
  full_name='CoreML.Specification.ArrayFeatureType.ShapeRange',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='sizeRanges', full_name='CoreML.Specification.ArrayFeatureType.ShapeRange.sizeRanges', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1258,
  serialized_end=1323,
)

_ARRAYFEATURETYPE = _descriptor.Descriptor(
  name='ArrayFeatureType',
  full_name='CoreML.Specification.ArrayFeatureType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='shape', full_name='CoreML.Specification.ArrayFeatureType.shape', index=0,
      number=1, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dataType', full_name='CoreML.Specification.ArrayFeatureType.dataType', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='enumeratedShapes', full_name='CoreML.Specification.ArrayFeatureType.enumeratedShapes', index=2,
      number=21, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='shapeRange', full_name='CoreML.Specification.ArrayFeatureType.shapeRange', index=3,
      number=31, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='intDefaultValue', full_name='CoreML.Specification.ArrayFeatureType.intDefaultValue', index=4,
      number=41, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='floatDefaultValue', full_name='CoreML.Specification.ArrayFeatureType.floatDefaultValue', index=5,
      number=51, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='doubleDefaultValue', full_name='CoreML.Specification.ArrayFeatureType.doubleDefaultValue', index=6,
      number=61, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_ARRAYFEATURETYPE_SHAPE, _ARRAYFEATURETYPE_ENUMERATEDSHAPES, _ARRAYFEATURETYPE_SHAPERANGE, ],
  enum_types=[
    _ARRAYFEATURETYPE_ARRAYDATATYPE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='ShapeFlexibility', full_name='CoreML.Specification.ArrayFeatureType.ShapeFlexibility',
      index=0, containing_type=None, fields=[]),
    _descriptor.OneofDescriptor(
      name='defaultOptionalValue', full_name='CoreML.Specification.ArrayFeatureType.defaultOptionalValue',
      index=1, containing_type=None, fields=[]),
  ],
  serialized_start=801,
  serialized_end=1455,
)


_DICTIONARYFEATURETYPE = _descriptor.Descriptor(
  name='DictionaryFeatureType',
  full_name='CoreML.Specification.DictionaryFeatureType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='int64KeyType', full_name='CoreML.Specification.DictionaryFeatureType.int64KeyType', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='stringKeyType', full_name='CoreML.Specification.DictionaryFeatureType.stringKeyType', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='KeyType', full_name='CoreML.Specification.DictionaryFeatureType.KeyType',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=1458,
  serialized_end=1622,
)


_SEQUENCEFEATURETYPE = _descriptor.Descriptor(
  name='SequenceFeatureType',
  full_name='CoreML.Specification.SequenceFeatureType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='int64Type', full_name='CoreML.Specification.SequenceFeatureType.int64Type', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='stringType', full_name='CoreML.Specification.SequenceFeatureType.stringType', index=1,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sizeRange', full_name='CoreML.Specification.SequenceFeatureType.sizeRange', index=2,
      number=101, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='Type', full_name='CoreML.Specification.SequenceFeatureType.Type',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=1625,
  serialized_end=1830,
)


_FEATURETYPE = _descriptor.Descriptor(
  name='FeatureType',
  full_name='CoreML.Specification.FeatureType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='int64Type', full_name='CoreML.Specification.FeatureType.int64Type', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='doubleType', full_name='CoreML.Specification.FeatureType.doubleType', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='stringType', full_name='CoreML.Specification.FeatureType.stringType', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='imageType', full_name='CoreML.Specification.FeatureType.imageType', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='multiArrayType', full_name='CoreML.Specification.FeatureType.multiArrayType', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dictionaryType', full_name='CoreML.Specification.FeatureType.dictionaryType', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sequenceType', full_name='CoreML.Specification.FeatureType.sequenceType', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='isOptional', full_name='CoreML.Specification.FeatureType.isOptional', index=7,
      number=1000, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='Type', full_name='CoreML.Specification.FeatureType.Type',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=1833,
  serialized_end=2327,
)

_IMAGEFEATURETYPE_IMAGESIZE.containing_type = _IMAGEFEATURETYPE
_IMAGEFEATURETYPE_ENUMERATEDIMAGESIZES.fields_by_name['sizes'].message_type = _IMAGEFEATURETYPE_IMAGESIZE
_IMAGEFEATURETYPE_ENUMERATEDIMAGESIZES.containing_type = _IMAGEFEATURETYPE
_IMAGEFEATURETYPE_IMAGESIZERANGE.fields_by_name['widthRange'].message_type = _SIZERANGE
_IMAGEFEATURETYPE_IMAGESIZERANGE.fields_by_name['heightRange'].message_type = _SIZERANGE
_IMAGEFEATURETYPE_IMAGESIZERANGE.containing_type = _IMAGEFEATURETYPE
_IMAGEFEATURETYPE.fields_by_name['enumeratedSizes'].message_type = _IMAGEFEATURETYPE_ENUMERATEDIMAGESIZES
_IMAGEFEATURETYPE.fields_by_name['imageSizeRange'].message_type = _IMAGEFEATURETYPE_IMAGESIZERANGE
_IMAGEFEATURETYPE.fields_by_name['colorSpace'].enum_type = _IMAGEFEATURETYPE_COLORSPACE
_IMAGEFEATURETYPE_COLORSPACE.containing_type = _IMAGEFEATURETYPE
_IMAGEFEATURETYPE.oneofs_by_name['SizeFlexibility'].fields.append(
  _IMAGEFEATURETYPE.fields_by_name['enumeratedSizes'])
_IMAGEFEATURETYPE.fields_by_name['enumeratedSizes'].containing_oneof = _IMAGEFEATURETYPE.oneofs_by_name['SizeFlexibility']
_IMAGEFEATURETYPE.oneofs_by_name['SizeFlexibility'].fields.append(
  _IMAGEFEATURETYPE.fields_by_name['imageSizeRange'])
_IMAGEFEATURETYPE.fields_by_name['imageSizeRange'].containing_oneof = _IMAGEFEATURETYPE.oneofs_by_name['SizeFlexibility']
_ARRAYFEATURETYPE_SHAPE.containing_type = _ARRAYFEATURETYPE
_ARRAYFEATURETYPE_ENUMERATEDSHAPES.fields_by_name['shapes'].message_type = _ARRAYFEATURETYPE_SHAPE
_ARRAYFEATURETYPE_ENUMERATEDSHAPES.containing_type = _ARRAYFEATURETYPE
_ARRAYFEATURETYPE_SHAPERANGE.fields_by_name['sizeRanges'].message_type = _SIZERANGE
_ARRAYFEATURETYPE_SHAPERANGE.containing_type = _ARRAYFEATURETYPE
_ARRAYFEATURETYPE.fields_by_name['dataType'].enum_type = _ARRAYFEATURETYPE_ARRAYDATATYPE
_ARRAYFEATURETYPE.fields_by_name['enumeratedShapes'].message_type = _ARRAYFEATURETYPE_ENUMERATEDSHAPES
_ARRAYFEATURETYPE.fields_by_name['shapeRange'].message_type = _ARRAYFEATURETYPE_SHAPERANGE
_ARRAYFEATURETYPE_ARRAYDATATYPE.containing_type = _ARRAYFEATURETYPE
_ARRAYFEATURETYPE.oneofs_by_name['ShapeFlexibility'].fields.append(
  _ARRAYFEATURETYPE.fields_by_name['enumeratedShapes'])
_ARRAYFEATURETYPE.fields_by_name['enumeratedShapes'].containing_oneof = _ARRAYFEATURETYPE.oneofs_by_name['ShapeFlexibility']
_ARRAYFEATURETYPE.oneofs_by_name['ShapeFlexibility'].fields.append(
  _ARRAYFEATURETYPE.fields_by_name['shapeRange'])
_ARRAYFEATURETYPE.fields_by_name['shapeRange'].containing_oneof = _ARRAYFEATURETYPE.oneofs_by_name['ShapeFlexibility']
_ARRAYFEATURETYPE.oneofs_by_name['defaultOptionalValue'].fields.append(
  _ARRAYFEATURETYPE.fields_by_name['intDefaultValue'])
_ARRAYFEATURETYPE.fields_by_name['intDefaultValue'].containing_oneof = _ARRAYFEATURETYPE.oneofs_by_name['defaultOptionalValue']
_ARRAYFEATURETYPE.oneofs_by_name['defaultOptionalValue'].fields.append(
  _ARRAYFEATURETYPE.fields_by_name['floatDefaultValue'])
_ARRAYFEATURETYPE.fields_by_name['floatDefaultValue'].containing_oneof = _ARRAYFEATURETYPE.oneofs_by_name['defaultOptionalValue']
_ARRAYFEATURETYPE.oneofs_by_name['defaultOptionalValue'].fields.append(
  _ARRAYFEATURETYPE.fields_by_name['doubleDefaultValue'])
_ARRAYFEATURETYPE.fields_by_name['doubleDefaultValue'].containing_oneof = _ARRAYFEATURETYPE.oneofs_by_name['defaultOptionalValue']
_DICTIONARYFEATURETYPE.fields_by_name['int64KeyType'].message_type = _INT64FEATURETYPE
_DICTIONARYFEATURETYPE.fields_by_name['stringKeyType'].message_type = _STRINGFEATURETYPE
_DICTIONARYFEATURETYPE.oneofs_by_name['KeyType'].fields.append(
  _DICTIONARYFEATURETYPE.fields_by_name['int64KeyType'])
_DICTIONARYFEATURETYPE.fields_by_name['int64KeyType'].containing_oneof = _DICTIONARYFEATURETYPE.oneofs_by_name['KeyType']
_DICTIONARYFEATURETYPE.oneofs_by_name['KeyType'].fields.append(
  _DICTIONARYFEATURETYPE.fields_by_name['stringKeyType'])
_DICTIONARYFEATURETYPE.fields_by_name['stringKeyType'].containing_oneof = _DICTIONARYFEATURETYPE.oneofs_by_name['KeyType']
_SEQUENCEFEATURETYPE.fields_by_name['int64Type'].message_type = _INT64FEATURETYPE
_SEQUENCEFEATURETYPE.fields_by_name['stringType'].message_type = _STRINGFEATURETYPE
_SEQUENCEFEATURETYPE.fields_by_name['sizeRange'].message_type = _SIZERANGE
_SEQUENCEFEATURETYPE.oneofs_by_name['Type'].fields.append(
  _SEQUENCEFEATURETYPE.fields_by_name['int64Type'])
_SEQUENCEFEATURETYPE.fields_by_name['int64Type'].containing_oneof = _SEQUENCEFEATURETYPE.oneofs_by_name['Type']
_SEQUENCEFEATURETYPE.oneofs_by_name['Type'].fields.append(
  _SEQUENCEFEATURETYPE.fields_by_name['stringType'])
_SEQUENCEFEATURETYPE.fields_by_name['stringType'].containing_oneof = _SEQUENCEFEATURETYPE.oneofs_by_name['Type']
_FEATURETYPE.fields_by_name['int64Type'].message_type = _INT64FEATURETYPE
_FEATURETYPE.fields_by_name['doubleType'].message_type = _DOUBLEFEATURETYPE
_FEATURETYPE.fields_by_name['stringType'].message_type = _STRINGFEATURETYPE
_FEATURETYPE.fields_by_name['imageType'].message_type = _IMAGEFEATURETYPE
_FEATURETYPE.fields_by_name['multiArrayType'].message_type = _ARRAYFEATURETYPE
_FEATURETYPE.fields_by_name['dictionaryType'].message_type = _DICTIONARYFEATURETYPE
_FEATURETYPE.fields_by_name['sequenceType'].message_type = _SEQUENCEFEATURETYPE
_FEATURETYPE.oneofs_by_name['Type'].fields.append(
  _FEATURETYPE.fields_by_name['int64Type'])
_FEATURETYPE.fields_by_name['int64Type'].containing_oneof = _FEATURETYPE.oneofs_by_name['Type']
_FEATURETYPE.oneofs_by_name['Type'].fields.append(
  _FEATURETYPE.fields_by_name['doubleType'])
_FEATURETYPE.fields_by_name['doubleType'].containing_oneof = _FEATURETYPE.oneofs_by_name['Type']
_FEATURETYPE.oneofs_by_name['Type'].fields.append(
  _FEATURETYPE.fields_by_name['stringType'])
_FEATURETYPE.fields_by_name['stringType'].containing_oneof = _FEATURETYPE.oneofs_by_name['Type']
_FEATURETYPE.oneofs_by_name['Type'].fields.append(
  _FEATURETYPE.fields_by_name['imageType'])
_FEATURETYPE.fields_by_name['imageType'].containing_oneof = _FEATURETYPE.oneofs_by_name['Type']
_FEATURETYPE.oneofs_by_name['Type'].fields.append(
  _FEATURETYPE.fields_by_name['multiArrayType'])
_FEATURETYPE.fields_by_name['multiArrayType'].containing_oneof = _FEATURETYPE.oneofs_by_name['Type']
_FEATURETYPE.oneofs_by_name['Type'].fields.append(
  _FEATURETYPE.fields_by_name['dictionaryType'])
_FEATURETYPE.fields_by_name['dictionaryType'].containing_oneof = _FEATURETYPE.oneofs_by_name['Type']
_FEATURETYPE.oneofs_by_name['Type'].fields.append(
  _FEATURETYPE.fields_by_name['sequenceType'])
_FEATURETYPE.fields_by_name['sequenceType'].containing_oneof = _FEATURETYPE.oneofs_by_name['Type']
DESCRIPTOR.message_types_by_name['Int64FeatureType'] = _INT64FEATURETYPE
DESCRIPTOR.message_types_by_name['DoubleFeatureType'] = _DOUBLEFEATURETYPE
DESCRIPTOR.message_types_by_name['StringFeatureType'] = _STRINGFEATURETYPE
DESCRIPTOR.message_types_by_name['SizeRange'] = _SIZERANGE
DESCRIPTOR.message_types_by_name['ImageFeatureType'] = _IMAGEFEATURETYPE
DESCRIPTOR.message_types_by_name['ArrayFeatureType'] = _ARRAYFEATURETYPE
DESCRIPTOR.message_types_by_name['DictionaryFeatureType'] = _DICTIONARYFEATURETYPE
DESCRIPTOR.message_types_by_name['SequenceFeatureType'] = _SEQUENCEFEATURETYPE
DESCRIPTOR.message_types_by_name['FeatureType'] = _FEATURETYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Int64FeatureType = _reflection.GeneratedProtocolMessageType('Int64FeatureType', (_message.Message,), dict(
  DESCRIPTOR = _INT64FEATURETYPE,
  __module__ = 'FeatureTypes_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.Int64FeatureType)
  ))
_sym_db.RegisterMessage(Int64FeatureType)

DoubleFeatureType = _reflection.GeneratedProtocolMessageType('DoubleFeatureType', (_message.Message,), dict(
  DESCRIPTOR = _DOUBLEFEATURETYPE,
  __module__ = 'FeatureTypes_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.DoubleFeatureType)
  ))
_sym_db.RegisterMessage(DoubleFeatureType)

StringFeatureType = _reflection.GeneratedProtocolMessageType('StringFeatureType', (_message.Message,), dict(
  DESCRIPTOR = _STRINGFEATURETYPE,
  __module__ = 'FeatureTypes_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.StringFeatureType)
  ))
_sym_db.RegisterMessage(StringFeatureType)

SizeRange = _reflection.GeneratedProtocolMessageType('SizeRange', (_message.Message,), dict(
  DESCRIPTOR = _SIZERANGE,
  __module__ = 'FeatureTypes_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.SizeRange)
  ))
_sym_db.RegisterMessage(SizeRange)

ImageFeatureType = _reflection.GeneratedProtocolMessageType('ImageFeatureType', (_message.Message,), dict(

  ImageSize = _reflection.GeneratedProtocolMessageType('ImageSize', (_message.Message,), dict(
    DESCRIPTOR = _IMAGEFEATURETYPE_IMAGESIZE,
    __module__ = 'FeatureTypes_pb2'
    # @@protoc_insertion_point(class_scope:CoreML.Specification.ImageFeatureType.ImageSize)
    ))
  ,

  EnumeratedImageSizes = _reflection.GeneratedProtocolMessageType('EnumeratedImageSizes', (_message.Message,), dict(
    DESCRIPTOR = _IMAGEFEATURETYPE_ENUMERATEDIMAGESIZES,
    __module__ = 'FeatureTypes_pb2'
    # @@protoc_insertion_point(class_scope:CoreML.Specification.ImageFeatureType.EnumeratedImageSizes)
    ))
  ,

  ImageSizeRange = _reflection.GeneratedProtocolMessageType('ImageSizeRange', (_message.Message,), dict(
    DESCRIPTOR = _IMAGEFEATURETYPE_IMAGESIZERANGE,
    __module__ = 'FeatureTypes_pb2'
    # @@protoc_insertion_point(class_scope:CoreML.Specification.ImageFeatureType.ImageSizeRange)
    ))
  ,
  DESCRIPTOR = _IMAGEFEATURETYPE,
  __module__ = 'FeatureTypes_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.ImageFeatureType)
  ))
_sym_db.RegisterMessage(ImageFeatureType)
_sym_db.RegisterMessage(ImageFeatureType.ImageSize)
_sym_db.RegisterMessage(ImageFeatureType.EnumeratedImageSizes)
_sym_db.RegisterMessage(ImageFeatureType.ImageSizeRange)

ArrayFeatureType = _reflection.GeneratedProtocolMessageType('ArrayFeatureType', (_message.Message,), dict(

  Shape = _reflection.GeneratedProtocolMessageType('Shape', (_message.Message,), dict(
    DESCRIPTOR = _ARRAYFEATURETYPE_SHAPE,
    __module__ = 'FeatureTypes_pb2'
    # @@protoc_insertion_point(class_scope:CoreML.Specification.ArrayFeatureType.Shape)
    ))
  ,

  EnumeratedShapes = _reflection.GeneratedProtocolMessageType('EnumeratedShapes', (_message.Message,), dict(
    DESCRIPTOR = _ARRAYFEATURETYPE_ENUMERATEDSHAPES,
    __module__ = 'FeatureTypes_pb2'
    # @@protoc_insertion_point(class_scope:CoreML.Specification.ArrayFeatureType.EnumeratedShapes)
    ))
  ,

  ShapeRange = _reflection.GeneratedProtocolMessageType('ShapeRange', (_message.Message,), dict(
    DESCRIPTOR = _ARRAYFEATURETYPE_SHAPERANGE,
    __module__ = 'FeatureTypes_pb2'
    # @@protoc_insertion_point(class_scope:CoreML.Specification.ArrayFeatureType.ShapeRange)
    ))
  ,
  DESCRIPTOR = _ARRAYFEATURETYPE,
  __module__ = 'FeatureTypes_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.ArrayFeatureType)
  ))
_sym_db.RegisterMessage(ArrayFeatureType)
_sym_db.RegisterMessage(ArrayFeatureType.Shape)
_sym_db.RegisterMessage(ArrayFeatureType.EnumeratedShapes)
_sym_db.RegisterMessage(ArrayFeatureType.ShapeRange)

DictionaryFeatureType = _reflection.GeneratedProtocolMessageType('DictionaryFeatureType', (_message.Message,), dict(
  DESCRIPTOR = _DICTIONARYFEATURETYPE,
  __module__ = 'FeatureTypes_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.DictionaryFeatureType)
  ))
_sym_db.RegisterMessage(DictionaryFeatureType)

SequenceFeatureType = _reflection.GeneratedProtocolMessageType('SequenceFeatureType', (_message.Message,), dict(
  DESCRIPTOR = _SEQUENCEFEATURETYPE,
  __module__ = 'FeatureTypes_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.SequenceFeatureType)
  ))
_sym_db.RegisterMessage(SequenceFeatureType)

FeatureType = _reflection.GeneratedProtocolMessageType('FeatureType', (_message.Message,), dict(
  DESCRIPTOR = _FEATURETYPE,
  __module__ = 'FeatureTypes_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.FeatureType)
  ))
_sym_db.RegisterMessage(FeatureType)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('H\003'))
# @@protoc_insertion_point(module_scope)



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import math

import coremltools.converters

from coremltools.converters.mil.mil import get_new_symbol
from coremltools.converters.mil.mil.types.symbolic import is_symbolic
from ._op_reqs import *


def broadcast_shapes(shape_x, shape_y):
    """
    Check and broadcast given input shapes.
    :param shape_x: tuple of int or symbols
        Shape of the first tensor (possibly symbolic).
    :param shape_y: tuple of int or symbols
        Shape of the second tensor (possibly symbolic).
    :return: tuple of int or symbols
        Result from broadcast.
    """
    shape_x = tuple(shape_x)
    shape_y = tuple(shape_y)
    if len(shape_x) < len(shape_y):
        shape_x = tuple([1] * (len(shape_y) - len(shape_x))) + shape_x
    if len(shape_y) < len(shape_x):
        shape_y = tuple([1] * (len(shape_x) - len(shape_y))) + shape_y

    ret_shapes = list()
    for i in range(len(shape_x)):
        x_unknown = is_symbolic(shape_x[i])
        y_unknown = is_symbolic(shape_y[i])
        if shape_x[i] == 1:
            ret_shapes.append(shape_y[i])
        elif shape_y[i] == 1:
            ret_shapes.append(shape_x[i])
        elif not y_unknown and shape_y[i] > 1:
            if not x_unknown and shape_x[i] != shape_y[i]:
                raise ValueError(
                    "Incompatible dim {} in shapes {} vs. {}".format(
                        i, shape_x, shape_y
                    )
                )
            ret_shapes.append(shape_y[i])
        elif not x_unknown and shape_x[i] > 1:
            if not y_unknown and shape_x[i] != shape_y[i]:
                raise ValueError(
                    "Incompatible dim {} in shapes {} vs. {}".format(
                        i, shape_x, shape_y
                    )
                )
            ret_shapes.append(shape_x[i])
        elif x_unknown or y_unknown:
            ret_shapes.append(get_new_symbol())
        else:
            assert shape_x[i] == shape_y[i]
            ret_shapes.append(shape_x[i])

    return tuple(ret_shapes)


def promoted_primitive_type(type1, type2):
    """
    Given a pair of tensor or primitive types, find the smallest type that can store an instance
    of their primitive type.
    """
    ptype1 = type1.get_primitive() if types.is_tensor(type1) else type1
    ptype2 = type2.get_primitive() if types.is_tensor(type2) else type2
    return types.promote_types(ptype1, ptype2)


def effective_kernel(kernel_shape, dilations):
    """

    Args:
        kernel_shape: tuple[int] representing the kernel shape in each
            given dimension.
        dilations: tuple[int] representing the dilation of the kernel
            in each given dimension.  Must be the same length as
            kernel_shape, and is assumed to give the dimensions in
            the same order as kernel_shape

    Returns: tuple[int] representing the effective shape of the kernel
        in each given dimension, with each dimension in the order given,
        taking into account dilation.
        See http://deeplearning.net/software/theano/tutorial/conv_arithmetic.html#dilated-convolutions
        Note that a dilation of 1 is equivalent to having no dilation.

    """
    if len(kernel_shape) != len(dilations):
        raise ValueError(
            "kernel_shape ({}) and dilations ({}) must be the same length".format(
                len(kernel_shape), len(dilations)
            )
        )
    return [(k - 1) * d + 1 for k, d in zip(kernel_shape, dilations)]


def aggregated_pad(
    pad_type,
    kernel_shape,
    input_shape=None,
    strides=None,
    dilations=None,
    custom_pad=None,
):
    """
    Args
        pad_type: string. Must be one of ('same', 'valid', 'custom')

        kernel_shape: [kH, kW, ...]: spatial kernel dims (excluding channels)

        input_shape: [iH, iW, ...]: spatial input dims (excluding channels)
            Required iff pad_type == 'same'

        strides: [sH, sW, ...]: spatial strides (excluding channels)
            Required iff pad_type == 'same'

        dilations: [dH, dW, ...]: dilations (excluding channels)
            If not provided, defaults to [1, 1, ...], effectively no dilation.

        custom_pad: Required iff pad_type == 'custom'.
            custom_pad[2*i], custom_pad[2*i+1] are before/after custom padding
            for spatial dim i.


    Returns:
        A list of total (before + after) padding for each spatial dimension in kernel_shape.
    """
    num_spatial_dims = len(kernel_shape)
    if dilations is None:
        dilations = [1] * num_spatial_dims
    elif len(dilations) != num_spatial_dims:
        raise ValueError(
            "dilations must have same length as kernel_shape ({}, but got {})".format(
                num_spatial_dims, len(dilations)
            )
        )
    if pad_type == "same":
        if input_shape is None or len(input_shape) != num_spatial_dims:
            raise ValueError(
                "For SAME padding input_shape must not be None and must have "
                "same length as kernel_shape ({}, but got {})".format(
                    num_spatial_dims,
                    len(input_shape) if input_shape is not None else "None",
                )
            )
        if strides is None or len(strides) != num_spatial_dims:
            raise ValueError(
                "For SAME padding strides must not be None and must have "
                "same length as kernel_shape ({}, but got {})".format(
                    num_spatial_dims, len(strides) if strides is not None else "None"
                )
            )
        effective_ks = effective_kernel(kernel_shape, dilations)
        return [
            int(max(0, s * math.ceil(float(i) / float(s)) - i + k - s))
            if not is_symbolic(i) else get_new_symbol()
            for i, k, s in zip(input_shape, effective_ks, strides)
        ]
    if pad_type == "valid":
        return [0] * num_spatial_dims
    if pad_type == "custom":
        if custom_pad is None or len(custom_pad) != 2 * num_spatial_dims:
            raise ValueError("Invalid custom_pad.")
        return [
            custom_pad[2 * d] + custom_pad[2 * d + 1] for d in range(num_spatial_dims)
        ]
    raise ValueError('Invalid padding pad_type "{}"'.format(pad_type))


def spatial_dimensions_out_shape(
    pad_type, input_shape, kernel_shape, strides, dilations=None, custom_pad=None
):
    """
    Args
        pad_type: string. Must be one of ('same', 'valid', 'custom')

        input_shape: [iH, iW, ...]: spatial input dims (excluding channels)
            Required iff pad_type == 'same'

        kernel_shape: [kH, kW, ...]: spatial kernel dims (excluding channels)

        strides: [sH, sW, ...]: spatial strides (excluding channels)
            Required iff pad_type == 'same'

        dilations: [dH, dW, ...]: dilations (excluding channels)
            If not provided, defaults to [1, 1, ...], effectively no dilation.

        custom_pad: Required iff pad_type == 'custom'.
            custom_pad[2*i], custom_pad[2*i+1] are before/after custom padding
            for spatial dim i.


    Returns:
        A list of spatial output sizes for each spatial dimension of kernel_shape.

    """
    num_spatial_dims = len(kernel_shape)
    if dilations is None:
        dilations = [1] * num_spatial_dims
    if custom_pad is None:
        custom_pad = [0] * num_spatial_dims * 2
    if not (
        len(input_shape)
        == len(kernel_shape)
        == len(strides)
        == len(dilations)
        == len(custom_pad) / 2
    ):
        raise ValueError(
            "input_shape (length {}), kernel_shape (length {}), "
            "strides (length {}), dilations (length {}), and "
            "custom_pad (length {}) divided by two must all be "
            "the same length".format(
                len(input_shape),
                len(kernel_shape),
                len(strides),
                len(dilations),
                len(custom_pad),
            )
        )

    pad = aggregated_pad(
        pad_type=pad_type,
        kernel_shape=kernel_shape,
        input_shape=input_shape,
        strides=strides,
        dilations=dilations,
        custom_pad=custom_pad,
    )
    effective_ks = effective_kernel(kernel_shape, dilations)
    out_shape = [
        (input_shape[r] + pad[r] - effective_ks[r]) // strides[r] + 1
        for r in range(num_spatial_dims)
    ]
    return [dim if not is_symbolic(dim) else get_new_symbol() for dim in out_shape]



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: TreeEnsemble.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import DataStructures_pb2 as DataStructures__pb2
try:
  FeatureTypes__pb2 = DataStructures__pb2.FeatureTypes__pb2
except AttributeError:
  FeatureTypes__pb2 = DataStructures__pb2.FeatureTypes_pb2

from .DataStructures_pb2 import *

DESCRIPTOR = _descriptor.FileDescriptor(
  name='TreeEnsemble.proto',
  package='CoreML.Specification',
  syntax='proto3',
  serialized_pb=_b('\n\x12TreeEnsemble.proto\x12\x14\x43oreML.Specification\x1a\x14\x44\x61taStructures.proto\"\xc4\x06\n\x16TreeEnsembleParameters\x12\x44\n\x05nodes\x18\x01 \x03(\x0b\x32\x35.CoreML.Specification.TreeEnsembleParameters.TreeNode\x12\x1f\n\x17numPredictionDimensions\x18\x02 \x01(\x04\x12\x1b\n\x13\x62\x61sePredictionValue\x18\x03 \x03(\x01\x1a\xa5\x05\n\x08TreeNode\x12\x0e\n\x06treeId\x18\x01 \x01(\x04\x12\x0e\n\x06nodeId\x18\x02 \x01(\x04\x12\\\n\x0cnodeBehavior\x18\x03 \x01(\x0e\x32\x46.CoreML.Specification.TreeEnsembleParameters.TreeNode.TreeNodeBehavior\x12\x1a\n\x12\x62ranchFeatureIndex\x18\n \x01(\x04\x12\x1a\n\x12\x62ranchFeatureValue\x18\x0b \x01(\x01\x12\x17\n\x0ftrueChildNodeId\x18\x0c \x01(\x04\x12\x18\n\x10\x66\x61lseChildNodeId\x18\r \x01(\x04\x12#\n\x1bmissingValueTracksTrueChild\x18\x0e \x01(\x08\x12\\\n\x0e\x65valuationInfo\x18\x14 \x03(\x0b\x32\x44.CoreML.Specification.TreeEnsembleParameters.TreeNode.EvaluationInfo\x12\x17\n\x0frelativeHitRate\x18\x1e \x01(\x01\x1a\x42\n\x0e\x45valuationInfo\x12\x17\n\x0f\x65valuationIndex\x18\x01 \x01(\x04\x12\x17\n\x0f\x65valuationValue\x18\x02 \x01(\x01\"\xcf\x01\n\x10TreeNodeBehavior\x12\x1e\n\x1a\x42ranchOnValueLessThanEqual\x10\x00\x12\x19\n\x15\x42ranchOnValueLessThan\x10\x01\x12!\n\x1d\x42ranchOnValueGreaterThanEqual\x10\x02\x12\x1c\n\x18\x42ranchOnValueGreaterThan\x10\x03\x12\x16\n\x12\x42ranchOnValueEqual\x10\x04\x12\x19\n\x15\x42ranchOnValueNotEqual\x10\x05\x12\x0c\n\x08LeafNode\x10\x06\"\xc7\x02\n\x16TreeEnsembleClassifier\x12\x42\n\x0ctreeEnsemble\x18\x01 \x01(\x0b\x32,.CoreML.Specification.TreeEnsembleParameters\x12Z\n\x17postEvaluationTransform\x18\x02 \x01(\x0e\x32\x39.CoreML.Specification.TreeEnsemblePostEvaluationTransform\x12?\n\x11stringClassLabels\x18\x64 \x01(\x0b\x32\".CoreML.Specification.StringVectorH\x00\x12=\n\x10int64ClassLabels\x18\x65 \x01(\x0b\x32!.CoreML.Specification.Int64VectorH\x00\x42\r\n\x0b\x43lassLabels\"\xb7\x01\n\x15TreeEnsembleRegressor\x12\x42\n\x0ctreeEnsemble\x18\x01 \x01(\x0b\x32,.CoreML.Specification.TreeEnsembleParameters\x12Z\n\x17postEvaluationTransform\x18\x02 \x01(\x0e\x32\x39.CoreML.Specification.TreeEnsemblePostEvaluationTransform*\x9d\x01\n#TreeEnsemblePostEvaluationTransform\x12\x0f\n\x0bNoTransform\x10\x00\x12\x1a\n\x16\x43lassification_SoftMax\x10\x01\x12\x17\n\x13Regression_Logistic\x10\x02\x12\x30\n,Classification_SoftMaxWithZeroClassReference\x10\x03\x42\x02H\x03P\x00\x62\x06proto3')
  ,
  dependencies=[DataStructures__pb2.DESCRIPTOR,],
  public_dependencies=[DataStructures__pb2.DESCRIPTOR,])

_TREEENSEMBLEPOSTEVALUATIONTRANSFORM = _descriptor.EnumDescriptor(
  name='TreeEnsemblePostEvaluationTransform',
  full_name='CoreML.Specification.TreeEnsemblePostEvaluationTransform',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NoTransform', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Classification_SoftMax', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Regression_Logistic', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Classification_SoftMaxWithZeroClassReference', index=3, number=3,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=1422,
  serialized_end=1579,
)
_sym_db.RegisterEnumDescriptor(_TREEENSEMBLEPOSTEVALUATIONTRANSFORM)

TreeEnsemblePostEvaluationTransform = enum_type_wrapper.EnumTypeWrapper(_TREEENSEMBLEPOSTEVALUATIONTRANSFORM)
NoTransform = 0
Classification_SoftMax = 1
Regression_Logistic = 2
Classification_SoftMaxWithZeroClassReference = 3


_TREEENSEMBLEPARAMETERS_TREENODE_TREENODEBEHAVIOR = _descriptor.EnumDescriptor(
  name='TreeNodeBehavior',
  full_name='CoreML.Specification.TreeEnsembleParameters.TreeNode.TreeNodeBehavior',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='BranchOnValueLessThanEqual', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BranchOnValueLessThan', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BranchOnValueGreaterThanEqual', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BranchOnValueGreaterThan', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BranchOnValueEqual', index=4, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BranchOnValueNotEqual', index=5, number=5,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LeafNode', index=6, number=6,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=696,
  serialized_end=903,
)
_sym_db.RegisterEnumDescriptor(_TREEENSEMBLEPARAMETERS_TREENODE_TREENODEBEHAVIOR)


_TREEENSEMBLEPARAMETERS_TREENODE_EVALUATIONINFO = _descriptor.Descriptor(
  name='EvaluationInfo',
  full_name='CoreML.Specification.TreeEnsembleParameters.TreeNode.EvaluationInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='evaluationIndex', full_name='CoreML.Specification.TreeEnsembleParameters.TreeNode.EvaluationInfo.evaluationIndex', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='evaluationValue', full_name='CoreML.Specification.TreeEnsembleParameters.TreeNode.EvaluationInfo.evaluationValue', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=627,
  serialized_end=693,
)

_TREEENSEMBLEPARAMETERS_TREENODE = _descriptor.Descriptor(
  name='TreeNode',
  full_name='CoreML.Specification.TreeEnsembleParameters.TreeNode',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='treeId', full_name='CoreML.Specification.TreeEnsembleParameters.TreeNode.treeId', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='nodeId', full_name='CoreML.Specification.TreeEnsembleParameters.TreeNode.nodeId', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='nodeBehavior', full_name='CoreML.Specification.TreeEnsembleParameters.TreeNode.nodeBehavior', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='branchFeatureIndex', full_name='CoreML.Specification.TreeEnsembleParameters.TreeNode.branchFeatureIndex', index=3,
      number=10, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='branchFeatureValue', full_name='CoreML.Specification.TreeEnsembleParameters.TreeNode.branchFeatureValue', index=4,
      number=11, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='trueChildNodeId', full_name='CoreML.Specification.TreeEnsembleParameters.TreeNode.trueChildNodeId', index=5,
      number=12, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='falseChildNodeId', full_name='CoreML.Specification.TreeEnsembleParameters.TreeNode.falseChildNodeId', index=6,
      number=13, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='missingValueTracksTrueChild', full_name='CoreML.Specification.TreeEnsembleParameters.TreeNode.missingValueTracksTrueChild', index=7,
      number=14, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='evaluationInfo', full_name='CoreML.Specification.TreeEnsembleParameters.TreeNode.evaluationInfo', index=8,
      number=20, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='relativeHitRate', full_name='CoreML.Specification.TreeEnsembleParameters.TreeNode.relativeHitRate', index=9,
      number=30, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_TREEENSEMBLEPARAMETERS_TREENODE_EVALUATIONINFO, ],
  enum_types=[
    _TREEENSEMBLEPARAMETERS_TREENODE_TREENODEBEHAVIOR,
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=226,
  serialized_end=903,
)

_TREEENSEMBLEPARAMETERS = _descriptor.Descriptor(
  name='TreeEnsembleParameters',
  full_name='CoreML.Specification.TreeEnsembleParameters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='nodes', full_name='CoreML.Specification.TreeEnsembleParameters.nodes', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='numPredictionDimensions', full_name='CoreML.Specification.TreeEnsembleParameters.numPredictionDimensions', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='basePredictionValue', full_name='CoreML.Specification.TreeEnsembleParameters.basePredictionValue', index=2,
      number=3, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_TREEENSEMBLEPARAMETERS_TREENODE, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=67,
  serialized_end=903,
)


_TREEENSEMBLECLASSIFIER = _descriptor.Descriptor(
  name='TreeEnsembleClassifier',
  full_name='CoreML.Specification.TreeEnsembleClassifier',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='treeEnsemble', full_name='CoreML.Specification.TreeEnsembleClassifier.treeEnsemble', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='postEvaluationTransform', full_name='CoreML.Specification.TreeEnsembleClassifier.postEvaluationTransform', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='stringClassLabels', full_name='CoreML.Specification.TreeEnsembleClassifier.stringClassLabels', index=2,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='int64ClassLabels', full_name='CoreML.Specification.TreeEnsembleClassifier.int64ClassLabels', index=3,
      number=101, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='ClassLabels', full_name='CoreML.Specification.TreeEnsembleClassifier.ClassLabels',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=906,
  serialized_end=1233,
)


_TREEENSEMBLEREGRESSOR = _descriptor.Descriptor(
  name='TreeEnsembleRegressor',
  full_name='CoreML.Specification.TreeEnsembleRegressor',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='treeEnsemble', full_name='CoreML.Specification.TreeEnsembleRegressor.treeEnsemble', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='postEvaluationTransform', full_name='CoreML.Specification.TreeEnsembleRegressor.postEvaluationTransform', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1236,
  serialized_end=1419,
)

_TREEENSEMBLEPARAMETERS_TREENODE_EVALUATIONINFO.containing_type = _TREEENSEMBLEPARAMETERS_TREENODE
_TREEENSEMBLEPARAMETERS_TREENODE.fields_by_name['nodeBehavior'].enum_type = _TREEENSEMBLEPARAMETERS_TREENODE_TREENODEBEHAVIOR
_TREEENSEMBLEPARAMETERS_TREENODE.fields_by_name['evaluationInfo'].message_type = _TREEENSEMBLEPARAMETERS_TREENODE_EVALUATIONINFO
_TREEENSEMBLEPARAMETERS_TREENODE.containing_type = _TREEENSEMBLEPARAMETERS
_TREEENSEMBLEPARAMETERS_TREENODE_TREENODEBEHAVIOR.containing_type = _TREEENSEMBLEPARAMETERS_TREENODE
_TREEENSEMBLEPARAMETERS.fields_by_name['nodes'].message_type = _TREEENSEMBLEPARAMETERS_TREENODE
_TREEENSEMBLECLASSIFIER.fields_by_name['treeEnsemble'].message_type = _TREEENSEMBLEPARAMETERS
_TREEENSEMBLECLASSIFIER.fields_by_name['postEvaluationTransform'].enum_type = _TREEENSEMBLEPOSTEVALUATIONTRANSFORM
_TREEENSEMBLECLASSIFIER.fields_by_name['stringClassLabels'].message_type = DataStructures__pb2._STRINGVECTOR
_TREEENSEMBLECLASSIFIER.fields_by_name['int64ClassLabels'].message_type = DataStructures__pb2._INT64VECTOR
_TREEENSEMBLECLASSIFIER.oneofs_by_name['ClassLabels'].fields.append(
  _TREEENSEMBLECLASSIFIER.fields_by_name['stringClassLabels'])
_TREEENSEMBLECLASSIFIER.fields_by_name['stringClassLabels'].containing_oneof = _TREEENSEMBLECLASSIFIER.oneofs_by_name['ClassLabels']
_TREEENSEMBLECLASSIFIER.oneofs_by_name['ClassLabels'].fields.append(
  _TREEENSEMBLECLASSIFIER.fields_by_name['int64ClassLabels'])
_TREEENSEMBLECLASSIFIER.fields_by_name['int64ClassLabels'].containing_oneof = _TREEENSEMBLECLASSIFIER.oneofs_by_name['ClassLabels']
_TREEENSEMBLEREGRESSOR.fields_by_name['treeEnsemble'].message_type = _TREEENSEMBLEPARAMETERS
_TREEENSEMBLEREGRESSOR.fields_by_name['postEvaluationTransform'].enum_type = _TREEENSEMBLEPOSTEVALUATIONTRANSFORM
DESCRIPTOR.message_types_by_name['TreeEnsembleParameters'] = _TREEENSEMBLEPARAMETERS
DESCRIPTOR.message_types_by_name['TreeEnsembleClassifier'] = _TREEENSEMBLECLASSIFIER
DESCRIPTOR.message_types_by_name['TreeEnsembleRegressor'] = _TREEENSEMBLEREGRESSOR
DESCRIPTOR.enum_types_by_name['TreeEnsemblePostEvaluationTransform'] = _TREEENSEMBLEPOSTEVALUATIONTRANSFORM
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

TreeEnsembleParameters = _reflection.GeneratedProtocolMessageType('TreeEnsembleParameters', (_message.Message,), dict(

  TreeNode = _reflection.GeneratedProtocolMessageType('TreeNode', (_message.Message,), dict(

    EvaluationInfo = _reflection.GeneratedProtocolMessageType('EvaluationInfo', (_message.Message,), dict(
      DESCRIPTOR = _TREEENSEMBLEPARAMETERS_TREENODE_EVALUATIONINFO,
      __module__ = 'TreeEnsemble_pb2'
      # @@protoc_insertion_point(class_scope:CoreML.Specification.TreeEnsembleParameters.TreeNode.EvaluationInfo)
      ))
    ,
    DESCRIPTOR = _TREEENSEMBLEPARAMETERS_TREENODE,
    __module__ = 'TreeEnsemble_pb2'
    # @@protoc_insertion_point(class_scope:CoreML.Specification.TreeEnsembleParameters.TreeNode)
    ))
  ,
  DESCRIPTOR = _TREEENSEMBLEPARAMETERS,
  __module__ = 'TreeEnsemble_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.TreeEnsembleParameters)
  ))
_sym_db.RegisterMessage(TreeEnsembleParameters)
_sym_db.RegisterMessage(TreeEnsembleParameters.TreeNode)
_sym_db.RegisterMessage(TreeEnsembleParameters.TreeNode.EvaluationInfo)

TreeEnsembleClassifier = _reflection.GeneratedProtocolMessageType('TreeEnsembleClassifier', (_message.Message,), dict(
  DESCRIPTOR = _TREEENSEMBLECLASSIFIER,
  __module__ = 'TreeEnsemble_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.TreeEnsembleClassifier)
  ))
_sym_db.RegisterMessage(TreeEnsembleClassifier)

TreeEnsembleRegressor = _reflection.GeneratedProtocolMessageType('TreeEnsembleRegressor', (_message.Message,), dict(
  DESCRIPTOR = _TREEENSEMBLEREGRESSOR,
  __module__ = 'TreeEnsemble_pb2'
  # @@protoc_insertion_point(class_scope:CoreML.Specification.TreeEnsembleRegressor)
  ))
_sym_db.RegisterMessage(TreeEnsembleRegressor)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('H\003'))
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-

#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from __future__ import print_function as _
from __future__ import division as _
from __future__ import absolute_import as _

from coremltools.converters.mil.mil.passes.pass_registry import register_pass
from coremltools.converters.mil.mil import Builder as mb
import numpy as np
from coremltools.converters.mil.mil.types.symbolic import is_symbolic, any_symbolic

def match_pattern(op):
    if op.outputs[0] in op.enclosing_block.outputs:
        return None

    if op.op_type == "concat":
        if op.interleave.val:
            return None

        # check that axis is -3 and rank is 4
        rank = op.values[0].rank
        if rank != 4:
            return None
        axis = op.axis.val
        if axis > 0:
            axis = axis - rank
        if axis != -3:
            return None

        # check that all inputs to concat have fully defined shapes
        for in_ in op.values:
            if any_symbolic(in_.shape):
                return None

        # check that all inputs to concat have the same shape
        inshape = list(op.values[0].shape)
        for v in op.values[1:]:
            for i in range(rank):
                if inshape[i] != v.shape[i]:
                    return None

        # check that this concat is connected to exactly 1 reshape op
        child_ops = list(op.outputs[0].child_ops)
        if len(child_ops) == 1:
            if list(child_ops)[0].op_type == "reshape":
                return op
    return None

def try_to_transform(concat_op, add_op, block):
    all_ops = [concat_op]
    B, C, H, W = list(concat_op.values[0].shape)
    n = len(concat_op.values)

    # check that reshape shapes the input to (B, n, C, H, W)
    reshape_op1 = concat_op.outputs[0].child_ops[0]
    reshape_shape1 = reshape_op1.shape.val
    if reshape_shape1 is None:
        return False
    if not isinstance(reshape_shape1, np.ndarray):
        return False
    reshape_shape1 = list(reshape_shape1)
    if reshape_shape1 != [B, n, C, H, W]:
        return False
    all_ops.append(reshape_op1)

    # check that after reshape is a transpose op with perm=[0, 2, 1, 3, 4]
    if len(list(reshape_op1.outputs[0].child_ops)) != 1:
        return False
    transpose_op = list(reshape_op1.outputs[0].child_ops)[0]
    if transpose_op.op_type != 'transpose':
        return False
    perm = transpose_op.perm.val
    if perm is None:
        return
    if list(perm) != [0, 2, 1, 3, 4]:
        return False
    all_ops.append(transpose_op)

    # check that after transpose is another reshape with [B, . , H, W]
    if len(list(transpose_op.outputs[0].child_ops)) != 1:
        return False
    reshape_op2 = list(transpose_op.outputs[0].child_ops)[0]
    if reshape_op2.op_type != 'reshape':
        return False
    reshape_shape2 = reshape_op2.shape.val
    if reshape_shape2 is None:
        return False
    if not isinstance(reshape_shape2, np.ndarray):
        return False
    reshape_shape2 = list(reshape_shape2)
    if len(reshape_shape2) != 4:
        return False
    if [reshape_shape2[0], reshape_shape2[-2], reshape_shape2[-1]] != [B, H, W]:
        return False
    all_ops.append(reshape_op2)

    # check that none of the op in this pattern is connected to the output
    # (except the last mul op)
    for i, op in enumerate(all_ops):
        if i == len(all_ops) - 1:
            continue
        for out in op.outputs:
            if out in block.outputs:
                return False

    # add a new concat op
    out_name = reshape_op2.outputs[0].name
    x = mb.concat(values=concat_op.values, axis=concat_op.axis.val, interleave=True, name=out_name, before_op=concat_op)

    reshape_op2.enclosing_block.replace_uses_of_var_after_op(
        anchor_op=reshape_op2, old_var=reshape_op2.outputs[0], new_var=x
    )

    # Remove all the ops at once
    block.remove_ops(all_ops)
    return True

def fuse_concat_interleave(block):
    fusion_status = False
    for op in list(block.operations):
        for b in op.blocks:
            block_changed = True
            while block_changed:
                block_changed = fuse_concat_interleave(b)
        if len(op.blocks) > 0:
            continue

        concat_op = match_pattern(op)
        if concat_op is not None:
            with block:
                fusion_status = try_to_transform(op, concat_op, block)
            # has to break as the downstream iterator is affected.
            if fusion_status:
                return fusion_status
    return fusion_status

@register_pass(namespace="common")
def detect_concat_interleave(prog):
    """
    Detect the pattern "concat-->reshape--->transpose--->reshape", where concat is
    along the channel axis (axis=-3), and map this pattern to the concat interleave op.

    This pattern occurs, for example, in the torchvision's shufflenet model.

    Given:
        %3 = concat(%1.a, %1.b, ..., axis=-3, interleave=False) #shape = (B, n*C, H, W)
        %4 = reshape(%3) #shape = (B, n, C, H, W)
        %5 = transpose(%4, perm=[0, 2, 1, 3, 4]) # shape = (B, C, n, H, W)
        %6 = reshape(%5) # shape = (B, C*n, H, W)

    Result:
        %6 = concat(%1.a, %1.b, ..., axis=-3, interleave=True)
    """
    for f_name, f in prog.functions.items():
        block_changed = True
        while block_changed:
            block_changed = fuse_concat_interleave(f)